from os import environ
from typing import Dict, Any, Optional, Union, List, Callable

from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent

from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkIfElseTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        enable: Optional[Union[bool, Callable[[DataFrame], bool]]] = None,
        enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = None,
        enable_sql: Optional[Union[str, Callable[[Optional[str]], str]]] = None,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        If enable flag is true then runs stages else runs else_stages


        :param enable: a boolean or a function that takes a DataFrame and returns a boolean
        :param enable_if_view_not_empty: enables if the view is not empty
        :param enable_sql: enables if sql returns any rows
        :param stages: list of transformers or a function that returns a list of transformers
        :param else_stages: list of transformers or a function that returns a list of transformers
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.enable: Optional[Union[bool, Callable[[DataFrame], bool]]] = enable

        self.enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = enable_if_view_not_empty

        self.enable_sql: Optional[Union[str, Callable[[Optional[str]], str]]] = (
            enable_sql
        )

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages
        self.else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = else_stages

        self.loop_id: Optional[str] = None

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        enable = self.enable(df) if callable(self.enable) else self.enable
        view_enable_if_view_not_empty = (
            self.enable_if_view_not_empty(self.loop_id)
            if callable(self.enable_if_view_not_empty)
            else self.enable_if_view_not_empty
        )
        enable_if_view_not_empty = (
            (
                df.sparkSession.catalog.tableExists(view_enable_if_view_not_empty)
                and not df.sparkSession.table(view_enable_if_view_not_empty).isEmpty()
            )
            if view_enable_if_view_not_empty
            else True
        )
        enable_sql = (
            self.enable_sql(self.loop_id)
            if callable(self.enable_sql)
            else self.enable_sql
        )
        enable_if_sql = df.sparkSession.sql(enable_sql) if enable_sql else True
        if (enable or enable is None) and enable_if_view_not_empty and enable_if_sql:
            stages: List[Transformer] = (
                self.stages if isinstance(self.stages, list) else self.stages()
            )
        else:
            if progress_logger is not None:
                progress_logger.write_to_log(
                    self.getName() or "FrameworkIfElseTransformer",
                    f"Skipping stages because enable {self.enable} or "
                    + f"enable_if_view_not_empty {self.enable_if_view_not_empty} or enable_sql {self.enable_sql} did not evaluate to True",
                )
            stages = (
                []
                if self.else_stages is None
                else (
                    self.else_stages
                    if isinstance(self.else_stages, list)
                    else self.else_stages()
                )
            )

        telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
            telemetry_parent=self.telemetry_parent or TelemetryParent.get_null_parent()
        ).create_telemetry_span_creator(log_level=environ.get("LOGLEVEL"))

        for stage in stages:
            if hasattr(stage, "getName"):
                # noinspection Mypy
                stage_name = stage.getName()
                if stage_name is not None:
                    stage_name = f"{stage_name} ({stage.__class__.__name__})"
                else:
                    stage_name = stage.__class__.__name__
            else:
                stage_name = stage.__class__.__name__

            telemetry_span: TelemetrySpanWrapper
            async with telemetry_span_creator.create_telemetry_span_async(
                name=stage_name, attributes={}, telemetry_parent=self.telemetry_parent
            ) as telemetry_span:

                if isinstance(stage, LoopIdMixin):
                    stage.set_loop_id(self.loop_id)
                if isinstance(stage, TelemetryParentMixin):
                    stage.set_telemetry_parent(
                        telemetry_parent=telemetry_span.create_child_telemetry_parent()
                    )

                try:
                    if hasattr(stage, "transform_async"):
                        df = await stage.transform_async(df)
                    else:
                        df = stage.transform(df)
                except Exception as e:
                    if len(e.args) >= 1:
                        # e.args = (e.args[0] + f" in stage {stage_name}") + e.args[1:]
                        e.args = (f"In Stage ({stage_name})", *e.args)
                    raise e
        return df

    def as_dict(self) -> Dict[str, Any]:
        return {
            **(super().as_dict()),
            "enable": self.enable,
            "enable_if_view_not_empty": self.enable_if_view_not_empty,
            "enable_sql": self.enable_sql,
            "stages": (
                [s.as_dict() for s in self.stages]  # type: ignore
                if not callable(self.stages)
                else str(self.stages)
            ),
            "else_stages": (
                [s.as_dict() for s in self.else_stages]  # type: ignore
                if self.else_stages and not callable(self.else_stages)
                else str(self.else_stages)
            ),
        }



from pathlib import Path
from typing import Any, Dict, Optional, Union, Callable

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.errors import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes


class FhirExporter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1,
        mode: str = FileWriteModes.MODE_ERROR,
        delta_lake_table: Optional[str] = None,
    ):
        """
        Converts a dataframe to FHIR JSON

        :param file_path: where to store the generated json
        :param view: where to read the source data frame
        :param mode: file write mode
        :param delta_lake_table: use delta lake format
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert mode in FileWriteModes.MODE_CHOICES

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)

        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        # limit: int = self.getLimit()

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        file_format: str = "delta" if delta_lake_table else "json"

        self.logger.info(
            f"---- Started exporting for {name or view} to {file_path} ------"
        )

        with ProgressLogMetric(
            name=f"{name or view}_fhir_exporter", progress_logger=progress_logger
        ):
            try:
                mode = self.getMode()
                file_path_text = str(file_path)
                if view:
                    df_view: DataFrame = df.sparkSession.table(view)
                    if not spark_is_data_frame_empty(df=df_view):
                        self.logger.info(f"---- Reading from view {view} ------")
                        assert not spark_is_data_frame_empty(df=df_view), view
                        df_view.write.format(file_format).mode(mode).save(
                            path=file_path_text
                        )
                    else:
                        self.logger.info(
                            f"---- Skipped exporting for {name or view} to {file_path} "
                            f"since the view {view} is empty ------"
                        )
                else:
                    assert not spark_is_data_frame_empty(df=df)
                    df.write.format(file_format).mode(mode).option(
                        "ignoreNullFields", True
                    ).save(path=file_path_text)

            except AnalysisException as e:
                self.logger.exception(
                    f"[{name or view}]File write failed to {file_path}"
                )
                raise e

        self.logger.info(
            f"---- Finished exporting for {name or view} to {file_path} ----"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)



import json

import pytest
from helix_fhir_client_sdk.responses.get.fhir_get_list_response import (
    FhirGetListResponse,
)
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_result import (
    GetBatchResult,
)


async def test_read_resources_and_errors_from_response_success() -> None:
    # Mock FHIR server response
    response = FhirGetListResponse(
        response_text=json.dumps(
            [
                {"resourceType": "Patient", "id": "1"},
                {"resourceType": "Patient", "id": "2"},
            ]
        ),
        status=200,
        request_id="test_request_id",
        url="http://fhir-server",
        error=None,
        access_token="abc",
        total_count=2,
        extra_context_to_return={"extra": "context"},
        resource_type="Patient",
        id_=None,
        response_headers=None,
        results_by_url=[],
        storage_mode=CompressedDictStorageMode(),
    )

    # Call the method
    result: GetBatchResult = (
        FhirReceiverProcessor.read_resources_and_errors_from_response(response=response)
    )

    # Assert the results
    assert len(result.resources) == 2
    assert result.resources[0] == '{"resourceType": "Patient", "id": "1"}'
    assert result.resources[1] == '{"resourceType": "Patient", "id": "2"}'
    assert len(result.errors) == 0


@pytest.mark.asyncio
async def test_read_resources_and_errors_from_response_with_errors() -> None:
    print()
    # Mock FHIR server response with an error
    response = FhirGetListResponse(
        response_text=json.dumps(
            [
                {"resourceType": "Patient", "id": "1"},
                {
                    "resourceType": "OperationOutcome",
                    "issue": [
                        {
                            "severity": "error",
                            "code": "processing",
                            "diagnostics": "Error message",
                        }
                    ],
                },
            ]
        ),
        status=200,
        request_id="test_request_id",
        url="http://fhir-server",
        error="Error message",
        access_token="abc",
        total_count=2,
        extra_context_to_return={"extra": "context"},
        resource_type="Patient",
        id_=None,
        response_headers=None,
        results_by_url=[],
        storage_mode=CompressedDictStorageMode(),
    )

    # Call the method
    result: GetBatchResult = (
        FhirReceiverProcessor.read_resources_and_errors_from_response(response=response)
    )

    # Assert the results
    assert len(result.resources) == 1
    assert result.resources[0] == '{"resourceType": "Patient", "id": "1"}'
    assert len(result.errors) == 1
    print(json.dumps(result.errors[0].to_dict()))
    assert result.errors[0].to_dict() == {
        "url": "http://fhir-server",
        "status_code": 200,
        "error_text": '{"resourceType": "OperationOutcome", "issue": [{"severity": "error", "code": "processing", "diagnostics": "Error message"}]}',
        "request_id": "test_request_id",
    }


@pytest.mark.asyncio
async def test_read_resources_and_errors_from_response_empty() -> None:
    # Mock FHIR server response with no resources
    response = FhirGetListResponse(
        response_text="",
        status=200,
        request_id="test_request_id",
        url="http://fhir-server",
        error="Error message",
        access_token="abc",
        total_count=2,
        extra_context_to_return={"extra": "context"},
        resource_type="Patient",
        id_=None,
        response_headers=None,
        results_by_url=[],
        storage_mode=CompressedDictStorageMode(),
    )

    # Call the method
    result: GetBatchResult = (
        FhirReceiverProcessor.read_resources_and_errors_from_response(response=response)
    )

    # Assert the results
    assert len(result.resources) == 0
    assert len(result.errors) == 0



import os
import shutil
import tempfile
from typing import Any, Optional
from urllib import parse as url_parse
from urllib import request as url_request
from zipfile import ZipFile, is_zipfile

import boto3
import requests


class ThrowOnErrorOpener(url_request.FancyURLopener):
    def http_error_default(
        self, url: str, fp: Any, errcode: int, errmsg: str, headers: Any
    ) -> Any:
        raise Exception(f"{errcode}: {errmsg}")


class FileDownloader:
    """
    Module to download files from urls
    Inspired from python wget module: https://github.com/steveeJ/python-wget
    """

    def __init__(
        self,
        url: str,
        download_path: str,
        extract_archives: Optional[bool] = False,
    ):
        self.url = url
        self.download_to_path = download_path
        self.extract_archives = extract_archives

    @staticmethod
    def check_if_path_exists(path: str) -> bool:
        if not os.path.exists(path):
            raise FileNotFoundError(f"File does not exists at this path: {path}")
        return True

    @staticmethod
    def get_filename_from_url(url: str) -> str:
        filename = os.path.basename(url_parse.urlparse(url).path)
        filename = filename.strip(" \n\t.")
        if len(filename) == 0:
            return None  # type: ignore
        return filename

    @staticmethod
    def rename_filename_if_exists(filename: str) -> str:
        """Expands name portion of filename with numeric ' (x)' suffix to
        return filename that doesn't exist already.
        """
        dirname = os.path.dirname(filename)
        name, ext = os.path.basename(filename).rsplit(".", 1)
        names = [x for x in os.listdir(dirname) if x.startswith(name)]
        names = [x.rsplit(".", 1)[0] for x in names]
        suffixes = [x.replace(name, "") for x in names]
        suffixes = [x[2:-1] for x in suffixes if x.startswith(" (") and x.endswith(")")]
        indexes = [int(x) for x in suffixes if set(x) <= set("0123456789")]
        idx = 1
        if indexes:
            idx += sorted(indexes)[-1]
        return os.path.join(dirname, f"{name} ({idx}).{ext}")

    def extract_zip_files(
        self,
        filename: str,
        path: str,
        out_path: Optional[str] = None,
    ) -> Optional[str]:
        """
        This function extracts the archive files and returns the corresponding file path.
        Returns null if the specified path does not exists
        :param filename: str: Name of the archive file
        :param path: str: File path of the archive file
        :param out_path: optional(str): Destination file path where the extracted files should be stored (default is same as path)
        :returns filename: str: Output file path
        """
        file_loc = os.path.join(path, filename)
        out_path = out_path or path

        if self.check_if_path_exists(path):
            if is_zipfile(file_loc):
                with ZipFile(file_loc, "r") as zip_ref:
                    zip_ref.extractall(out_path)
                return out_path

        return None

    def download_files_locally(self, filename: str) -> str:
        local_file_protocol_length: int = len("file://")
        download_to_path = self.download_to_path[local_file_protocol_length:]
        prefix = (filename or download_to_path or "") + ".."
        (fd, tmpfile) = tempfile.mkstemp(".tmp", prefix=prefix, dir=".")
        os.close(fd)
        os.unlink(tmpfile)

        # Define callbacks in this code block
        # callback = None

        (tmpfile, headers) = ThrowOnErrorOpener().retrieve(self.url, tmpfile)

        if os.path.isdir(download_to_path):
            file_path = filename
            file_path = os.path.join(download_to_path, file_path)
        else:
            file_path = download_to_path or filename
        if os.path.exists(file_path):
            file_path = self.rename_filename_if_exists(file_path)
        shutil.move(tmpfile, file_path)

        if self.extract_archives:
            file_path = (
                self.extract_zip_files(
                    filename=file_path,
                    path=download_to_path,
                )
                or file_path
            )

        return file_path

    def download_files_to_s3(
        self, filename: str, bucket: str, download_path: str
    ) -> str:
        file_request = requests.get(self.url, stream=True)
        s3_resource = boto3.resource("s3")

        bucket = s3_resource.Bucket(bucket)
        s3_location = f"{download_path}/{filename}"

        if bucket.creation_date is not None:  # type: ignore
            return bucket.upload_fileobj(file_request.raw, s3_location)  # type: ignore
        else:
            raise s3_resource.exception.NoSuchBucket(f"{bucket} does not exists!!")

    def download_files_from_url(self) -> Optional[str]:
        """
        Function to download files from a url

        """
        filename = self.get_filename_from_url(self.url)
        url_segments = url_parse.urlparse(self.download_to_path)

        if url_segments.scheme in ["s3", "s3a"]:
            file_path = os.path.join(
                self.url,
                self.download_files_to_s3(
                    str(filename),
                    str(url_segments.netloc),
                    str(url_segments.path).strip("/"),
                ),
            )
        elif url_segments.scheme in ["file"]:
            file_path = self.download_files_locally(filename)
        else:
            raise NotImplementedError

        if file_path:
            return filename

        return None



from aioresponses import aioresponses
from pyspark.sql import SparkSession

from spark_pipeline_framework.transformers.fhir_receiver.v1.fhir_receiver_helpers import (
    FhirReceiverHelpers,
    GetBatchResult,
)


def test_get_batch_results_paging(spark_session: SparkSession) -> None:
    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "1", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=1",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "2", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_getpagesoffset=0&id%253Aabove=2",
            status=404,
        )

        result = FhirReceiverHelpers.get_batch_result(
            last_updated_after=None,
            last_updated_before=None,
            limit=10,
            page_size=5,
            server_url="http://fhir-server",
            log_level="DEBUG",
            action=None,
            action_payload=None,
            additional_parameters=None,
            filter_by_resource=None,
            filter_parameter=None,
            sort_fields=None,
            auth_server_url="http://fhir-server/token",
            auth_client_id=None,
            auth_client_secret=None,
            auth_login_token=None,
            auth_scopes=None,
            include_only_properties=None,
            separate_bundle_resources=False,
            expand_fhir_bundle=False,
            accept_type=None,
            content_type=None,
            additional_request_headers=None,
            accept_encoding=None,
            retry_count=None,
            exclude_status_codes_from_retry=None,
            auth_access_token=None,
            error_view=None,
            use_data_streaming=None,
            graph_json=None,
            ignore_status_codes=[],
            refresh_token_function=None,
            resource_name="Patient",
        )
        assert isinstance(result, GetBatchResult)
        assert len(result.resources) == 2
        assert result.resources[0] == '{"id": "1", "resourceType": "Patient"}'
        assert result.resources[1] == '{"id": "2", "resourceType": "Patient"}'


def test_get_batch_results_empty_bundle(spark_session: SparkSession) -> None:
    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient?_elements=id%2Csubject%2CrelatesTo&_count=5&_getpagesoffset=0",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "1", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_elements=id%252Csubject%252CrelatesTo&_getpagesoffset=0&id%253Aabove=1",
            payload={
                "resourceType": "Bundle",
                "entry": [{"resource": {"id": "2", "resourceType": "Patient"}}],
            },
        )
        m.get(
            "http://fhir-server/Patient?_count=5&_elements=id%252Csubject%252CrelatesTo&_getpagesoffset=0&id%253Aabove=2",
            payload={
                "resourceType": "Bundle",
                "type": "searchset",
                "timestamp": "2022-08-08T07:50:38.3838Z",
                "total": 0,
                "entry": [],
            },
        )

        result = FhirReceiverHelpers.get_batch_result(
            last_updated_after=None,
            last_updated_before=None,
            limit=5,
            page_size=None,
            server_url="http://fhir-server",
            log_level="DEBUG",
            action=None,
            action_payload=None,
            additional_parameters=None,
            filter_by_resource=None,
            filter_parameter=None,
            sort_fields=None,
            auth_server_url="http://fhir-server/token",
            auth_client_id=None,
            auth_client_secret=None,
            auth_login_token=None,
            auth_scopes=None,
            include_only_properties=["id,subject,relatesTo"],
            separate_bundle_resources=False,
            expand_fhir_bundle=True,
            accept_type=None,
            content_type=None,
            additional_request_headers=None,
            accept_encoding=None,
            retry_count=None,
            exclude_status_codes_from_retry=None,
            auth_access_token=None,
            error_view=None,
            use_data_streaming=None,
            graph_json=None,
            ignore_status_codes=[],
            refresh_token_function=None,
            resource_name="Patient",
        )
        assert isinstance(result, GetBatchResult)
        assert len(result.resources) == 2
        assert result.resources[0] == '{"id": "1", "resourceType": "Patient"}'
        assert result.resources[1] == '{"id": "2", "resourceType": "Patient"}'



import json
import logging
from datetime import datetime
from logging import Logger
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Callable,
    Iterator,
    AsyncGenerator,
)

import pandas as pd
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.responses.fhir_update_response import FhirUpdateResponse

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender_parameters import (
    FhirSenderParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_dataframe_udf import (
    AsyncPandasDataFrameUDF,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_item import (
    FhirMergeResponseItem,
)
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender_helpers_async import (
    update_json_bundle_to_fhir_async,
    send_fhir_delete_async,
    send_json_bundle_to_fhir_async,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.utilities.json_helpers import convert_dict_to_fhir_json
from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


class FhirSenderProcessor:
    # noinspection PyUnusedLocal
    @staticmethod
    async def process_chunk(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[FhirSenderParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Process a partition of data asynchronously

        :param run_context: run context
        :param input_values: input values
        :param parameters: parameters
        :param additional_parameters: additional parameters
        :return: output values
        """
        register()
        assert parameters
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        spark_partition_information: SparkPartitionInformation = (
            SparkPartitionInformation.from_current_task_context(
                chunk_index=run_context.chunk_index,
            )
        )
        message: str = f"FhirSenderProcessor:process_partition"
        # Format the time to include hours, minutes, seconds, and milliseconds
        formatted_time = datetime.now().strftime("%H:%M:%S.%f")[:-3]
        formatted_message: str = (
            f"{formatted_time}: "
            f"{message}"
            f" | Partition: {run_context.partition_index}"
            f" | Chunk: {run_context.chunk_index}"
            f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
            f" | {spark_partition_information}"
        )
        logger.info(formatted_message)

        count: int = 0
        try:
            r: FhirMergeResponse | FhirUpdateResponse | FhirDeleteResponse
            async for r in FhirSenderProcessor.send_partition_to_server_async(
                partition_index=run_context.partition_index,
                chunk_index=run_context.chunk_index,
                parameters=parameters,
                rows=input_values,
            ):
                logger.debug(
                    f"Received result"
                    f" | Partition: {run_context.partition_index}"
                    f" | Chunk: {run_context.chunk_index}"
                    f" | Status: {r.status}"
                    f" | Url: {r.url}"
                    f" | Count: {len(r.responses)}"
                )
                if isinstance(r, FhirMergeResponse):
                    for merge_response in FhirMergeResponseItem.from_merge_response(
                        merge_response=r
                    ):
                        count += 1
                        yield merge_response.to_dict()
                elif isinstance(r, FhirUpdateResponse):
                    count += 1
                    yield FhirMergeResponseItem.from_update_response(
                        update_response=r
                    ).to_dict()
                elif isinstance(r, FhirDeleteResponse):
                    count += 1  # no responses in delete
                    yield FhirMergeResponseItem.from_delete_response(
                        delete_response=r
                    ).to_dict()
        except Exception as e:
            logger.error(
                f"Error processing partition {run_context.partition_index} chunk {run_context.chunk_index}: {str(e)}"
            )
            # if an exception is thrown then return an error for each row
            for input_value in input_values:
                count += 1
                # print(
                #     f"FhirSenderProcessor.process_partition exception input_value: {input_value}"
                # )
                yield FhirMergeResponseItem.from_error(
                    e=e, resource_type=parameters.resource_name
                ).to_dict()
        # we actually want to error here since something strange happened
        assert count == len(input_values), f"count={count}, len={len(input_values)}"

    @staticmethod
    def get_process_partition_function(
        *, parameters: FhirSenderParameters
    ) -> Callable[[Iterable[pd.DataFrame]], Iterator[pd.DataFrame]]:
        """
        Returns a function that includes the passed parameters so that function
        can be used in a pandas_udf

        :param parameters: FhirSenderParameters
        :return: pandas_udf
        """

        return AsyncPandasDataFrameUDF(
            async_func=FhirSenderProcessor.process_chunk,  # type: ignore[arg-type]
            parameters=parameters,
            pandas_udf_parameters=AsyncPandasUdfParameters(
                max_chunk_size=parameters.batch_size or 100
            ),
        ).get_pandas_udf()

    @staticmethod
    # function that is called for each partition
    async def send_partition_to_server_async(
        *,
        partition_index: int,
        chunk_index: int,
        rows: Iterable[Dict[str, Any]],
        parameters: FhirSenderParameters,
    ) -> AsyncGenerator[
        FhirMergeResponse | FhirUpdateResponse | FhirDeleteResponse, None
    ]:
        """
        This function processes a partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        :param partition_index: index of partition
        :param chunk_index: index of chunk
        :param rows: rows to process
        :param parameters: parameters for this function
        :return: response from the server as FhirMergeResponseItem dicts
        """
        json_data_list: List[Dict[str, Any]] = list(rows)
        try:
            assert isinstance(json_data_list, list)
            if len(json_data_list) > 0:
                assert isinstance(json_data_list[0], dict)
            logger: Logger = get_logger(
                __name__,
                level=(
                    parameters.log_level
                    if parameters and parameters.log_level
                    else "INFO"
                ),
            )
            assert parameters
            assert isinstance(parameters, FhirSenderParameters)
            assert parameters.server_url
            assert isinstance(
                partition_index, int
            ), f"partition_index should be an int but is {type(partition_index)}"

            if len(json_data_list) == 0:
                assert len(json_data_list) > 0, "json_data_list should not be empty"
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    f"Sending batch (DEBUG) "
                    f" | Partition: {partition_index}/{parameters.total_partitions}"
                    f" | Chunk: {chunk_index}"
                    f" | Rows: {len(json_data_list)}"
                    f" | Operation: {parameters.operation}"
                    f" | Url: {parameters.server_url}/{parameters.resource_name}"
                    f" | data= {json_data_list}"
                )
            else:
                logger.info(
                    f"Sending batch "
                    f" | Partition: {partition_index}/{parameters.total_partitions}"
                    f" | Chunk: {chunk_index}"
                    f" | Rows: {len(json_data_list)}"
                    f" | Operation: {parameters.operation}"
                    f" | Url: {parameters.server_url}/{parameters.resource_name}"
                )

            request_id_list: List[str] = []
            responses: List[Dict[str, Any]] = []
            if FhirSenderOperation.operation_equals(
                parameters.operation, FhirSenderOperation.FHIR_OPERATION_PATCH
            ):
                for item in json_data_list:
                    item_value = json.loads(item["value"])
                    payload = [
                        json.loads(convert_dict_to_fhir_json(payload_item))
                        for payload_item in item_value["payload"]
                    ]
                    patch: FhirUpdateResponse
                    async for patch_result in update_json_bundle_to_fhir_async(
                        obj_id=item_value["id"],
                        json_data=json.dumps(payload),
                        server_url=parameters.server_url,
                        operation=parameters.operation,
                        validation_server_url=parameters.validation_server_url,
                        resource=parameters.resource_name,
                        logger=logger,
                        auth_server_url=parameters.auth_server_url,
                        auth_client_id=parameters.auth_client_id,
                        auth_client_secret=parameters.auth_client_secret,
                        auth_login_token=parameters.auth_login_token,
                        auth_scopes=parameters.auth_scopes,
                        auth_access_token=parameters.auth_access_token,
                        auth_well_known_url=parameters.auth_well_known_url,
                        additional_request_headers=parameters.additional_request_headers,
                        log_level=parameters.log_level,
                    ):
                        yield patch_result
            elif FhirSenderOperation.operation_equals(
                parameters.operation, FhirSenderOperation.FHIR_OPERATION_PUT
            ):
                for item in json_data_list:
                    item_value = json.loads(item["value"])
                    put_result: FhirUpdateResponse
                    async for put_result in update_json_bundle_to_fhir_async(
                        obj_id=item_value["id"],
                        json_data=convert_dict_to_fhir_json(item_value),
                        server_url=parameters.server_url,
                        operation=parameters.operation,
                        validation_server_url=parameters.validation_server_url,
                        resource=parameters.resource_name,
                        logger=logger,
                        auth_server_url=parameters.auth_server_url,
                        auth_client_id=parameters.auth_client_id,
                        auth_client_secret=parameters.auth_client_secret,
                        auth_login_token=parameters.auth_login_token,
                        auth_scopes=parameters.auth_scopes,
                        auth_access_token=parameters.auth_access_token,
                        auth_well_known_url=parameters.auth_well_known_url,
                        additional_request_headers=parameters.additional_request_headers,
                        log_level=parameters.log_level,
                    ):
                        yield put_result
            elif FhirSenderOperation.operation_equals(
                parameters.operation, FhirSenderOperation.FHIR_OPERATION_DELETE
            ):
                # FHIR doesn't support bulk deletes, so we have to send one at a time
                for item in json_data_list:
                    delete_result: FhirDeleteResponse
                    async for delete_result in send_fhir_delete_async(
                        obj_id=(item if "id" in item else json.loads(item["value"]))[
                            "id"
                        ],  # parse the JSON and extract the id
                        server_url=parameters.server_url,
                        resource=parameters.resource_name,
                        logger=logger,
                        auth_server_url=parameters.auth_server_url,
                        auth_client_id=parameters.auth_client_id,
                        auth_client_secret=parameters.auth_client_secret,
                        auth_login_token=parameters.auth_login_token,
                        auth_scopes=parameters.auth_scopes,
                        auth_access_token=parameters.auth_access_token,
                        auth_well_known_url=parameters.auth_well_known_url,
                        additional_request_headers=parameters.additional_request_headers,
                        log_level=parameters.log_level,
                    ):
                        yield delete_result
            elif FhirSenderOperation.operation_equals(
                parameters.operation, FhirSenderOperation.FHIR_OPERATION_MERGE
            ):
                if parameters.batch_size == 1:
                    # ensure we call one at a time. Partitioning does not guarantee that each
                    #   partition will have exactly the batch size
                    auth_access_token1: Optional[str] = parameters.auth_access_token
                    i: int = 0
                    count: int = len(json_data_list)
                    for item in json_data_list:
                        i += 1
                        current_bundle: List[Dict[str, Any]]
                        json_data_list_1: List[str] = [
                            (
                                convert_dict_to_fhir_json(item)
                                if "id" in item
                                else item["value"]
                            )
                        ]
                        first_item: Optional[Dict[str, Any]] = (
                            json.loads(json_data_list_1[0])
                            if len(json_data_list_1) > 0
                            else None
                        )
                        first_id = (
                            first_item.get("id") if first_item is not None else None
                        )
                        logger.debug(
                            f"Sending {i} of {count} from partition {partition_index} "
                            f"for {parameters.resource_name}: {first_id}"
                        )
                        if first_id:
                            result: Optional[FhirMergeResponse]
                            async for result in send_json_bundle_to_fhir_async(
                                id_=first_id,
                                json_data_list=json_data_list_1,
                                server_url=parameters.server_url,
                                validation_server_url=parameters.validation_server_url,
                                resource=parameters.resource_name,
                                logger=logger,
                                auth_server_url=parameters.auth_server_url,
                                auth_client_id=parameters.auth_client_id,
                                auth_client_secret=parameters.auth_client_secret,
                                auth_login_token=parameters.auth_login_token,
                                auth_scopes=parameters.auth_scopes,
                                auth_access_token=auth_access_token1,
                                auth_well_known_url=parameters.auth_well_known_url,
                                additional_request_headers=parameters.additional_request_headers,
                                log_level=parameters.log_level,
                                retry_count=parameters.retry_count,
                                exclude_status_codes_from_retry=parameters.exclude_status_codes_from_retry,
                                smart_merge=parameters.smart_merge,
                            ):
                                if result:
                                    auth_access_token1 = result.access_token
                                    if result.request_id:
                                        request_id_list.append(result.request_id)
                                    responses.extend(result.responses)
                                    yield result
                else:
                    # send a whole batch to the server at once
                    json_data_list_1 = [
                        (
                            convert_dict_to_fhir_json(item)
                            if "id" in item
                            else item["value"]
                        )
                        for item in json_data_list
                    ]
                    first_item = (
                        json.loads(json_data_list_1[0])
                        if len(json_data_list_1) > 0
                        else None
                    )
                    first_id = first_item.get("id") if first_item is not None else None
                    if first_id:
                        async for result in send_json_bundle_to_fhir_async(
                            id_=first_id,
                            json_data_list=json_data_list_1,
                            server_url=parameters.server_url,
                            validation_server_url=parameters.validation_server_url,
                            resource=parameters.resource_name,
                            auth_server_url=parameters.auth_server_url,
                            auth_client_id=parameters.auth_client_id,
                            auth_client_secret=parameters.auth_client_secret,
                            auth_login_token=parameters.auth_login_token,
                            auth_scopes=parameters.auth_scopes,
                            auth_access_token=parameters.auth_access_token,
                            auth_well_known_url=parameters.auth_well_known_url,
                            additional_request_headers=parameters.additional_request_headers,
                            logger=logger,
                            log_level=parameters.log_level,
                            retry_count=parameters.retry_count,
                            exclude_status_codes_from_retry=parameters.exclude_status_codes_from_retry,
                            smart_merge=parameters.smart_merge,
                        ):
                            if result and result.request_id:
                                request_id_list.append(result.request_id)
                            yield result
        except Exception as e:
            yield FhirMergeResponse(
                request_id="",
                status=400,
                error=str(e),
                json_data=json.dumps(json_data_list),
                access_token=parameters.auth_access_token,
                url=parameters.server_url,
                responses=[],
            )



from logging import Logger
from typing import Optional

from pyspark.sql import DataFrame

from spark_pipeline_framework.utilities.spark_execution_information.v1.spark_execution_information import (
    SparkExecutionInformation,
)


class PartitionCalculator:
    @staticmethod
    def calculate_ideal_partitions(
        *,
        executor_cores: Optional[int],
        executor_count: Optional[int],
        executor_memory: Optional[str],
        input_row_count: Optional[int],
        input_row_size: Optional[int],
        maximum_number_of_partitions: Optional[int],
        output_row_size: Optional[int],
        partition_size: Optional[int],
        percentage_of_memory_to_use: float,
        df: DataFrame,
        logger: Optional[Logger],
    ) -> Optional[int]:
        """
        Calculate the ideal number of partitions based on the input parameters.

        :param executor_cores: The number of cores per executor.
        :param executor_count: The number of executors.
        :param executor_memory: The memory available to each executor.
        :param input_row_count: The number of rows in the input DataFrame.
        :param input_row_size: The size of each row in the input DataFrame.
        :param maximum_number_of_partitions: The maximum number of partitions to use.
        :param output_row_size: The size of each row in the output DataFrame.
        :param partition_size: The size of each partition.
        :param percentage_of_memory_to_use: The percentage of memory to use.
        :param df: The DataFrame to use to calculate the ideal number of partitions.
        :param logger: The logger to use.
        :return: The ideal number of partitions if we were able to calculate it, otherwise None.
        """
        calculated_partitions: Optional[int] = None
        spark_execution_information: SparkExecutionInformation = (
            SparkExecutionInformation(
                df=df,
                executor_count=executor_count,
                executor_cores=executor_cores,
                executor_memory=executor_memory,
            )
        )
        # assume we can use only half of the executor memory
        executor_memory_available: Optional[int] = (
            spark_execution_information.current_executor_memory // 2
            if spark_execution_information.current_executor_memory
            else None
        )
        # if we have calculated executor instances then we can calculate the partitions
        if spark_execution_information.current_executor_instances is not None:
            # calculate the size available per executor
            size_available_per_executor: Optional[int] = (
                partition_size if partition_size else executor_memory_available
            )
            assert (
                size_available_per_executor is not None
            ), "partition_size is not set and spark.executor.memory config is also not set"
            # use only a percentage of the memory
            size_available_per_executor = int(
                size_available_per_executor * percentage_of_memory_to_use
            )
            # calculate estimated row size if input row size is not provided
            estimated_row_size: int = (
                len(str(df.rdd.first())) if not input_row_size else input_row_size
            )
            # if output row size is provided then add it to the input row size else double input row size
            if output_row_size is not None:
                estimated_row_size += output_row_size
            else:
                estimated_row_size += estimated_row_size  # assume output row size will be same as input size
            # calculate number of rows if not provided
            num_rows: int = df.count() if not input_row_count else input_row_count
            # calculate total size of the dataframe
            estimated_total_size: int = num_rows * estimated_row_size
            # now calculate the partitions as maximum of number of executors and the total size of the dataframe
            # but make sure we don't get more partitions than the number of rows
            calculated_partitions = max(
                spark_execution_information.current_executor_instances,
                int(estimated_total_size // size_available_per_executor),
            )
            # partitions should not be more than the number of rows
            calculated_partitions = min(num_rows, calculated_partitions)
            # partitions should not be more than the maximum number of partitions if passed
            if maximum_number_of_partitions is not None:
                calculated_partitions = min(
                    maximum_number_of_partitions, calculated_partitions
                )
            # Log the calculated partitions
            if logger:
                logger.info(
                    f"Calculated Partitions: {calculated_partitions}"
                    f" | Rows: {num_rows}"
                    f" | Estimated row size: {spark_execution_information.convert_bytes_to_human_readable(estimated_row_size)}"
                    f" | Executor Memory To Use: {spark_execution_information.convert_bytes_to_human_readable(size_available_per_executor)}"
                    f" | Executors: {spark_execution_information.current_executor_instances}"
                    f" | Cores: {spark_execution_information.current_executor_cores}"
                    f" | Executor Memory: {spark_execution_information.convert_bytes_to_human_readable(spark_execution_information.current_executor_memory)}"
                    f" | Executor Memory Available: {spark_execution_information.convert_bytes_to_human_readable(executor_memory_available)}"
                    + (
                        f" | AWS Instance Type: {spark_execution_information.spark_worker_node_type}"
                        if spark_execution_information.spark_worker_node_type
                        else ""
                    )
                    + (
                        f" | Cluster Target Workers: {spark_execution_information.spark_cluster_target_workers_count}"
                        if spark_execution_information.spark_cluster_target_workers_count
                        else ""
                    )
                )
        else:
            if logger:
                logger.warning(
                    "Could not calculate partitions automatically as `spark.executor.instances` config is not set"
                    " and no executor_count is provided"
                )
        return calculated_partitions



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import lit, to_json, col

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_fhir_meta_updater.v1.framework_fhir_meta_updater import (
    FrameworkFhirMetaUpdater,
)


def test_framework_fhir_meta_updater(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = spark_session.read.format("json").load(
        str(data_dir.joinpath("patient.json"))
    )
    df = df.withColumn("my_slug", lit("cigna"))
    df = df.withColumn(
        "my_url", lit("https://p-hi2.digitaledge.cigna.com/PatientAccess/v1-devportal")
    )
    df.createOrReplaceTempView("patients")

    # Act
    with ProgressLogger() as progress_logger:
        FrameworkFhirMetaUpdater(
            resource_type="Patient",
            source_view="patients",
            view="fixed_patients",
            slug_column="my_slug",
            url_column="my_url",
            connection_type="proa",
            progress_logger=progress_logger,
        ).transform(df)

    # Assert
    result_df = spark_session.table("fixed_patients")
    result_df.show(truncate=False)
    # result_df.write.format("json").save(str(data_dir.joinpath("expected_patient.json")))

    meta = json.loads(result_df.select(to_json(col("meta"))).collect()[0][0])
    print(meta)
    print(type(meta))

    assert meta == {
        "source": "https://p-hi2.digitaledge.cigna.com/PatientAccess/v1-devportal/Patient/cigna-0000000000000015116",
        "security": [
            {"system": "https://www.icanbwell.com/owner", "code": "cigna"},
            {"system": "https://www.icanbwell.com/access", "code": "cigna"},
            {"system": "https://www.icanbwell.com/vendor", "code": "cigna"},
            {"system": "https://www.icanbwell.com/connectionType", "code": "proa"},
        ],
    }



from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import cast

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType
from spark_fhir_schemas.r4.resources.patient import PatientSchema

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_exporter.v1.fhir_exporter import (
    FhirExporter,
)


def test_fhir_exporter_on_delta(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    fhir_data = data_dir.joinpath("fhir_calls").joinpath("patient")
    df: DataFrame = (
        spark_session.read.schema(cast(StructType, PatientSchema.get_schema()))
        .format("json")
        .option("multiLine", True)
        .load(str(fhir_data))
    )

    test_files_dir_delta = temp_folder.joinpath("test_files/patients")
    # convert to delta
    df.write.format("delta").save(str(test_files_dir_delta))

    df = spark_session.read.format("delta").load(str(test_files_dir_delta))

    df.show(truncate=False)

    df.createOrReplaceTempView("source")

    # Act
    with ProgressLogger() as progress_logger:
        FhirExporter(
            progress_logger=progress_logger,
            view="source",
            file_path=str(temp_folder.joinpath("out_delta")),
            delta_lake_table="table1",
        ).transform(df)

    # Assert
    result_df: DataFrame = (
        spark_session.read.schema(cast(StructType, PatientSchema.get_schema()))
        .format("json")
        .option("multiLine", True)
        .load(str(temp_folder.joinpath("out_delta")))
    )

    assert result_df.count() == 2



from pyspark.sql.session import SparkSession

from spark_pipeline_framework.transformers.framework_jdbc_exporter.v1.framework_jdbc_exporter import (
    FrameworkJdbcExporter,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes


def test_can_save_via_jdbc(spark_session: SparkSession) -> None:
    """
    Because testing against a database is a pain and most of the core transformation logic is in the base exporter,
    we're only testing that the options and format are correctly being exposed.
    """
    # Arrange
    view = "my_view"
    jdbc_url = "jdbc:mysql:user@password:host/db:port"
    table = "my_view_table"
    driver = "org.driver.FakeDriver"

    # Act
    exporter = FrameworkJdbcExporter(
        view=view,
        jdbc_url=jdbc_url,
        table=table,
        driver=driver,
        mode=FileWriteModes.MODE_OVERWRITE,
    )

    # Assert
    options = exporter.getOptions()
    assert options["url"] == jdbc_url
    assert options["driver"] == driver
    assert exporter.getFormat() == "jdbc"
    assert exporter.getMode() == "overwrite"


def test_can_specify_additional_writer_options(spark_session: SparkSession) -> None:
    """
    Because testing against a database is a pain and most of the core transformation logic is in the base exporter,
    we're only testing that the options and format are correctly being exposed.
    """
    # Arrange
    view = "my_view"
    jdbc_url = "jdbc:mysql:user@password:host/db:port"
    table = "my_view_table"
    driver = "org.driver.FakeDriver"
    column_types = "Column1 BIGINT, Column2 VARCHAR(1024), Column3 TEXT"
    options = {"customTableColumntypes": column_types}

    # Act
    exporter = FrameworkJdbcExporter(
        jdbc_url=jdbc_url,
        view=view,
        table=table,
        driver=driver,
        mode=FileWriteModes.MODE_OVERWRITE,
        options=options,
    )

    # Assert
    options = exporter.getOptions()
    assert options["url"] == jdbc_url
    assert options["driver"] == driver
    assert exporter.getFormat() == "jdbc"
    assert exporter.getMode() == "overwrite"
    assert options["customTableColumntypes"] == column_types



import abc
from abc import ABCMeta
from typing import List, Type

import structlog


from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor_factory import (
    StandardizingVendorFactory,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)

logger = structlog.get_logger(__file__)


class CacheHandler(metaclass=ABCMeta):
    """
    read vendor specific addresses form cache and return standard addresses
    """

    @abc.abstractmethod
    async def check_cache(self, raw_addresses: List[RawAddress]) -> CacheResult:
        """
        returns {'found': <list of addresses found in cache>, 'not_found': <list of addresses not found in cache>}
        """

    @abc.abstractmethod
    async def save_to_cache(
        self, vendor_responses: List[VendorResponse[BaseVendorApiResponse]]
    ) -> None:
        pass

    def _get_vendor_class(
        self, vendor_name: str
    ) -> Type[StandardizingVendor[BaseVendorApiResponse]]:
        return StandardizingVendorFactory.get_vendor_class(vendor_name)



import json
import os
from os import environ
from pathlib import Path
from shutil import rmtree
from typing import Any, Dict

from mockserver_client.mockserver_client import (
    MockServerFriendlyClient,
    mock_request,
    mock_response,
    times,
)
from pyspark.sql import SparkSession, DataFrame

from library.features.patients_fhir.v1.features_patients_fhir_v1 import (
    FeaturesPatientsFhirV1,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.send_automapper_to_fhir.v2.automapper_to_fhir_transformer import (
    AutoMapperToFhirTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


async def test_automapper_to_fhir_transformer_async(
    spark_session: SparkSession,
) -> None:
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    test_name = "automapper_to_fhir_transformer_test"

    environ["LOGLEVEL"] = "DEBUG"

    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_dir = data_dir.joinpath("temp")
    if os.path.isdir(temp_dir):
        rmtree(temp_dir)
    os.makedirs(temp_dir)

    def get_fhir_path(view: str, resource_name: str) -> str:
        return str(temp_dir.joinpath("output").joinpath(f"{resource_name}-{view}"))

    def get_fhir_response_path(view: str, resource_name: str) -> str:
        return str(
            temp_dir.joinpath("output").joinpath(f"{resource_name}-{view}-response")
        )

    # create the input dataframe
    patients_df: DataFrame = spark_session.createDataFrame(
        [  # type: ignore
            {
                "member_id": "1",
                "last_name": "Doe",
                "first_name": "John",
            },
            {
                "member_id": "2",
                "last_name": "Smith",
                "first_name": "Jane",
            },
        ]
    )
    patients_df.createOrReplaceTempView("patients")

    server_url = f"http://mock-server:1080/{test_name}"
    mock_client = MockServerFriendlyClient("http://mock-server:1080")
    mock_client.clear(test_name)
    mock_client.reset()

    expected_json = [
        {
            "resourceType": "Patient",
            "id": "1",
            "name": [{"use": "usual", "family": "Doe", "given": ["John"]}],
        },
        {
            "resourceType": "Patient",
            "id": "2",
            "name": [{"use": "usual", "family": "Smith", "given": ["Jane"]}],
        },
    ]

    merge_response = [
        {
            "resourceType": "Patient",
            "id": "1",
            "created": True,
        },
        {
            "resourceType": "Patient",
            "id": "2",
            "created": True,
        },
    ]

    mock_client.expect(
        request=mock_request(
            path=f"/{test_name}/Patient/1/$merge",
            body=json.dumps(expected_json),
            method="POST",
        ),
        response=mock_response(code=200, body=json.dumps(merge_response)),
        timing=times(1),
        file_path="1",
    )

    parameters: Dict[str, Any] = {}
    with ProgressLogger() as progress_logger:
        transformer = AutoMapperToFhirTransformer(
            name="AutoMapperToFhirTransformer",
            parameters=parameters,
            progress_logger=progress_logger,
            transformer=FeaturesPatientsFhirV1(
                parameters={
                    "source_view": "patients",
                    "view": "fhir_patients",
                },
                progress_logger=progress_logger,
            ),
            func_get_path=get_fhir_path,
            func_get_response_path=get_fhir_response_path,
            fhir_server_url=server_url,
            source_entity_name="Patient",
        )
        result_df = await transformer.transform_async(df)

        result_df.show()

        assert result_df.count() == 0

        mock_client.verify_expectations(
            test_name=test_name,
        )



from typing import Dict, Any, Optional

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql import Column
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import explode, regexp_extract

from spark_fhir_schemas.r4.resources.bundle import BundleSchema

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FhirResourceLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        file_path: str,
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert file_path
        assert view

        # add params
        self.file_path: Param[str] = Param(self, "file_path", "")
        self._setDefault(file_path=file_path)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        file_path: str = self.getFilePath()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        if progress_logger:
            progress_logger.write_to_log(
                name="FhirResourceLoader", message="Loading json file(s)..."
            )

        schema = BundleSchema.get_schema()
        df_entry: DataFrame = (
            df.sparkSession.read.option("multiLine", True)
            .json(path=str(file_path), schema=schema)  # type: ignore
            .select("entry")
        )

        df_resources: DataFrame = df_entry.select(
            explode(df_entry.entry.resource).alias("resource")
        )

        regex_output: Column = regexp_extract(
            str=df_resources.resource, pattern=r'."resourceType":"(\w+)"', idx=1
        )

        df_resources = df_resources.withColumn(colName="resourceType", col=regex_output)

        if progress_logger:
            progress_logger.write_to_log(
                name="FhirResourceLoader",
                message=f"Resources row count: {df_resources.count()}",
            )

        df_resources.createOrReplaceTempView(name=view)

        return df_resources

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> str:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)



from pathlib import Path
from typing import Dict, Any, cast, List

from pyspark.ml import Transformer

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from library.features.carriers_fhir.v1.features_carriers_fhir_v1 import (
    FeaturesCarriersFhirV1,
)
from library.features.carriers_python.v1.features_carriers_python_v1 import (
    FeaturesCarriersPythonV1,
)
from spark_pipeline_framework.pipelines.v2.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_drop_views_transformer.v1.framework_drop_views_transformer import (
    FrameworkDropViewsTransformer,
)
from spark_pipeline_framework.transformers.framework_if_else_transformer.v1.framework_if_else_transformer import (
    FrameworkIfElseTransformer,
)
from spark_pipeline_framework.transformers.framework_json_exporter.v1.framework_json_exporter import (
    FrameworkJsonExporter,
)
from spark_pipeline_framework.transformers.send_automapper_to_fhir.v1.automapper_to_fhir_transformer import (
    AutoMapperToFhirTransformer,
)


class SimplePipeline(FrameworkPipeline):
    def get_fhir_path(self, view: str, resource_name: str) -> str:
        return str(
            self.data_dir.joinpath("temp")
            .joinpath("output")
            .joinpath(f"{resource_name}-{view}")
        )

    def get_fhir_response_path(self, view: str, resource_name: str) -> str:
        return str(
            self.data_dir.joinpath("temp")
            .joinpath("output")
            .joinpath(f"{resource_name}-{view}-response")
        )

    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(SimplePipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="12345678",
            client_name="client_foo",
            vendor_name="vendor_foo",
        )
        self.data_dir: Path = Path(__file__).parent.joinpath("./")

        self.transformers = self.create_steps(
            cast(
                List[Transformer],
                [
                    FrameworkDropViewsTransformer(
                        name="",
                        parameters=parameters,
                        progress_logger=progress_logger,
                        views=["foo"],
                    ),
                    FrameworkCsvLoader(
                        name="FrameworkCsvLoader",
                        view="flights",
                        file_path=parameters["flights_path"],
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                    FeaturesCarriersV1(
                        parameters=parameters, progress_logger=progress_logger
                    ),
                    FrameworkIfElseTransformer(
                        enable=True,
                        progress_logger=progress_logger,
                        stages=[
                            FeaturesCarriersPythonV1(
                                parameters=parameters, progress_logger=progress_logger
                            ),
                        ],
                    ),
                    AutoMapperToFhirTransformer(
                        name="AutoMapperToFhirTransformer",
                        parameters=parameters,
                        progress_logger=progress_logger,
                        transformer=FeaturesCarriersFhirV1(
                            parameters=parameters, progress_logger=progress_logger
                        ),
                        func_get_path=self.get_fhir_path,
                        func_get_response_path=self.get_fhir_response_path,
                        fhir_server_url="http://mock-server:1080",
                        source_entity_name="members",
                    ),
                    FrameworkJsonExporter(
                        file_path=parameters["export_path"],
                        view="flights",
                        name="FrameworkJsonExporter",
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                ],
            )
        )



from typing import Callable

from pyspark.sql import DataFrame
from pyspark.errors import AnalysisException
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


def nonzero_view_rowcount(view: str) -> Callable[[DataFrame], bool]:
    def inner(df: DataFrame) -> bool:
        try:
            return not spark_is_data_frame_empty(df.sparkSession.table(view))
        except AnalysisException:
            return False

    return inner


def same_rowcount_two_views(view1: str, view2: str) -> Callable[[DataFrame], bool]:
    def inner(df: DataFrame) -> bool:
        try:
            return (
                df.sparkSession.table(view1).count()
                == df.sparkSession.table(view2).count()
            )
        except AnalysisException:
            return False

    return inner



import json
from os import environ, path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import Any, List, Dict, Optional

import pytest

from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.utilities.fhir_helper import FhirHelper
from helix_fhir_client_sdk.utilities.practitioner_generator import PractitionerGenerator
from pyspark.sql import SparkSession, DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver import (
    FhirReceiver,
)
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


@pytest.mark.parametrize("run_synchronously", [True])
@pytest.mark.parametrize("use_data_streaming", [True])
async def test_async_real_fhir_server_get_graph_error(
    spark_session: SparkSession, run_synchronously: bool, use_data_streaming: bool
) -> None:
    print()
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    patient_json_path: Path = temp_folder.joinpath("patient.json")
    resource_type = "Practitioner"
    async with FhirServerTestContext(
        resource_type=resource_type
    ) as fhir_server_test_context:

        environ["LOGLEVEL"] = "DEBUG"

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url

        fhir_client = await fhir_server_test_context.create_fhir_client_async()
        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        count: int = 2
        roles_per_practitioner: int = 2

        id_dict: Dict[str, List[str]] = PractitionerGenerator.get_ids(
            count=count, roles_per_practitioner=roles_per_practitioner
        )
        # delete Practitioner resources
        print(f"Deleting {count} {resource_type} resources: {id_dict[resource_type]}")
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type=resource_type,
            id_list=id_dict[resource_type],
        )
        # delete Organization resources
        print(f"Deleting {count} Organization resources: {id_dict['Organization']}")
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type="Organization",
            id_list=id_dict["Organization"],
        )
        # delete PractitionerRole resources
        print(
            f"Deleting {count} PractitionerRole resources: {id_dict['PractitionerRole']}"
        )
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type="PractitionerRole",
            id_list=id_dict["PractitionerRole"],
        )
        print("Finished deleting resources")

        fhir_client = await fhir_server_test_context.create_fhir_client_async()
        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        bundle: Dict[str, Any] = PractitionerGenerator.generate_resources_bundle(
            count=count, roles_per_practitioner=roles_per_practitioner
        )

        expected_resource_count = len(bundle["entry"])
        print("expected_resource_count", expected_resource_count)

        print(f"Merging bundle with {expected_resource_count} resources")
        merge_response: Optional[FhirMergeResponse] = (
            await FhirMergeResponse.from_async_generator(
                fhir_client.merge_async(json_data_list=[json.dumps(bundle)])
            )
        )
        assert merge_response is not None
        print(f"Merged {expected_resource_count} resources")
        print(json.dumps(merge_response.responses))
        assert merge_response.status == 200, merge_response.responses
        assert (
            len(merge_response.responses) == expected_resource_count
        ), merge_response.responses
        # assert merge_response.responses[0]["created"] is True, merge_response.responses

        bad_practitioner_graph = {
            "resourceType": "GraphDefinition",
            "id": "o",
            "name": "provider_slots",
            "status": "active",
            "start": "Foobar",
            "link": ["foo"],
        }
        # act
        df: DataFrame = create_empty_dataframe(spark_session=spark_session)

        id_df = spark_session.createDataFrame(
            [(s,) for s in id_dict[resource_type]], ["id"]
        )

        id_df.createOrReplaceTempView("id_view")

        parameters = {"flow_name": "Test Pipeline V2", "team_name": "Data Operations"}

        with ProgressLogger() as progress_logger:
            await FhirReceiver(
                server_url=fhir_server_url,
                resource=resource_type,
                action="$graph",
                id_view="id_view",
                additional_parameters=["contained=true"],
                separate_bundle_resources=True,
                action_payload=bad_practitioner_graph,
                file_path=patient_json_path,
                progress_logger=progress_logger,
                parameters=parameters,
                use_data_streaming=use_data_streaming,
                run_synchronously=run_synchronously,
                auth_well_known_url=auth_well_known_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                error_view="error_view",
            ).transform_async(df)

        # Assert
        json_df: DataFrame = df.sparkSession.read.json(str(patient_json_path))
        print("------- Result DataFrame -------")
        json_df.show(truncate=False)
        # json_df.printSchema()

        assert json_df.count() == 2

        # error_df: DataFrame = df.sparkSession.table("error_view")
        # print("------- Error DataFrame -------")
        # error_df.show(truncate=False)
        #
        # assert error_df.count() == 1
        # first_row = error_df.first()
        # assert first_row is not None
        # assert first_row["url"] is not None
        # assert (
        #     first_row["url"] == "http://fhir:3000/4_0_0/Practitioner/$graph?contained=true"
        # )
        # assert error_df.first().error == "The server could not process the request"



import logging

import pytest
from aioresponses import aioresponses
from compressedfhir.fhir.fhir_resource_list import FhirResourceList
from helix_fhir_client_sdk.responses.fhir_get_response import FhirGetResponse

from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)


@pytest.mark.asyncio
async def test_send_fhir_request_async() -> None:
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        result: FhirGetResponse
        async for result in FhirReceiverProcessor.send_fhir_request_async(
            logger=logging.getLogger(__name__),
            resource_id="1",
            server_url="http://fhir-server",
            parameters=parameters,
        ):
            assert isinstance(result, FhirGetResponse)
            resources = result.get_resources()
            assert isinstance(resources, FhirResourceList)
            assert [r.dict() for r in resources] == [
                {"resourceType": "Patient", "id": "1"}
            ]


@pytest.mark.asyncio
async def test_send_simple_fhir_request_async() -> None:
    parameters = get_fhir_receiver_parameters()

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        result: FhirGetResponse
        async for result in FhirReceiverProcessor.send_simple_fhir_request_async(
            id_="1",
            server_url="http://fhir-server",
            server_url_="http://fhir-server",
            parameters=parameters,
        ):
            assert isinstance(result, FhirGetResponse)
            resources = result.get_resources()
            assert isinstance(resources, FhirResourceList)
            assert [r.dict() for r in resources] == [
                {"resourceType": "Patient", "id": "1"}
            ]



from typing import Dict, Any, List, Set, Optional


class FhirMetaDataParser:
    """
    Parses FHIR metadata


    """

    @staticmethod
    def get_resource_types_from_metadata(metadata: Dict[str, Any]) -> List[str]:
        """
        Gets resource types present in the metadata


        :param metadata:
        """
        resource_types: Set[str] = set()
        rest_entry: Dict[str, Any]
        for rest_entry in metadata.get("rest", []):
            for resource in rest_entry.get("resource", []):
                resource_type: Optional[str] = resource.get("type")
                if resource_type is not None:
                    resource_types.add(resource_type)

        return list(resource_types)

    @staticmethod
    def get_fir_version_from_metadata(metadata: Dict[str, Any]) -> Optional[str]:
        """
        Gets FHIR version from metadata


        :param metadata:
        """
        fhir_version: Optional[str] = metadata.get("fhirVersion")
        return fhir_version



import json
from os import environ
from typing import Any, Callable, Dict, List, Optional, Union


from spark_pipeline_framework.transformers.athena_table_creator.v1.athena_table_creator import (
    AthenaTableCreator,
)
from spark_pipeline_framework.transformers.fhir_exporter.v1.fhir_exporter import (
    FhirExporter,
)
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender import FhirSender
from spark_pipeline_framework.transformers.send_automapper_to_fhir.exceptions.automapper_to_fhir_transformer_exception import (
    AutomapperToFhirTransformerException,
)
from spark_pipeline_framework.utilities.athena.athena_source_file_type import (
    AthenaSourceFileType,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.proxy_base import ProxyBase
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.transformers.framework_parquet_exporter.v1.framework_parquet_exporter import (
    FrameworkParquetExporter,
)

from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class AutoMapperToFhirTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        transformer: ProxyBase,
        func_get_path: Callable[[str, str], str],
        func_get_response_path: Callable[[str, str], str],
        fhir_server_url: str,
        source_entity_name: str,
        file_format: str = "json",
        fhir_validation_url: Optional[str] = None,
        athena_schema: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        send_to_fhir: Optional[bool] = True,
        mode: str = FileWriteModes.MODE_ERROR,
        run_synchronously: Optional[bool] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        sort_data: Optional[Dict[str, Dict[str, Any]]] = None,
        partition_by_column_name: Optional[str] = None,
        enable_repartitioning: bool = True,
        operation: Union[
            FhirSenderOperation, str
        ] = FhirSenderOperation.FHIR_OPERATION_MERGE.value,
        log_level: Optional[str] = None,
        smart_merge: Optional[bool] = None,
    ):
        """
        Runs the auto-mappers, saves the result to Athena db and then sends the results to fhir server

        :param transformer: transformer to run
        :param func_get_path: function that gets the local path for a resource
        :param func_get_response_path: function that gets the path to save the responses from the FHIR server
        :param fhir_server_url: url to FHIR server
        :param fhir_validation_url: url to FHIR validation server
        :param source_entity_name: resource name
        :param athena_schema: schema to use when writing data to Athena
        :param run_synchronously: (Optional) Run on the Spark master to make debugging easier on dev machines
        :param additional_request_headers: (Optional) Additional request headers to use
                                            (Eg: {"Accept-Charset": "utf-8"})
        :param sort_data: (Optional) Whether to sort the data. Format - {
                            KEY - view: view name,
                            VALUE - {
                                sort_by_column_name_and_type: "columnName and columnType to be used for sorting",
                                drop_fields_from_json: (list) List of fields to drop from json,
                                partition_by_column_name: (str) Name of the column that will be used to repartition df
                            }
                        }
        :param enable_repartitioning: Enable repartitioning or not, default True
        :param operation: The API operation to perform, such as merge, put, delete etc
        :param log_level: (Optional) Log level to use
        :param smart_merge: (Optional) Enable smart merge functionality
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert fhir_server_url

        self.logger = get_logger(__name__)
        self.logger.info(
            f"Parameters={json.dumps({str(key): str(value) for key, value in parameters.items()}) if parameters else 'None'}"
        )

        # add a param
        self.transformer: Param[ProxyBase] = Param(self, "transformer", "")
        self._setDefault(transformer=transformer)

        self.func_get_path: Param[Callable[[str, str], str]] = Param(
            self, "func_get_path", ""
        )
        self._setDefault(func_get_path=func_get_path)

        self.func_get_response_path: Param[Callable[[str, str], str]] = Param(
            self, "func_get_response_path", ""
        )
        self._setDefault(func_get_response_path=func_get_response_path)

        self.fhir_server_url: Param[str] = Param(self, "fhir_server_url", "")
        self._setDefault(fhir_server_url=fhir_server_url)

        self.file_format: Param[str] = Param(self, "file_format", "")
        self._setDefault(file_format=file_format)

        self.fhir_validation_url: Param[Optional[str]] = Param(
            self, "fhir_validation_url", ""
        )
        self._setDefault(fhir_validation_url=fhir_validation_url)

        self.source_entity_name: Param[str] = Param(self, "source_entity_name", "")
        self._setDefault(source_entity_name=source_entity_name)

        self.athena_schema: Param[Optional[str]] = Param(self, "athena_schema", "")
        self._setDefault(athena_schema=athena_schema)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[Optional[List[str]]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.send_to_fhir: Param[Optional[bool]] = Param(self, "send_to_fhir", "")
        self._setDefault(send_to_fhir=send_to_fhir)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.additional_request_headers: Param[Optional[Dict[str, str]]] = Param(
            self, "additional_request_headers", ""
        )
        self._setDefault(additional_request_headers=additional_request_headers)

        self.sort_data: Param[Optional[Dict[str, Dict[str, Any]]]] = Param(
            self, "sort_data", ""
        )
        self._setDefault(sort_data=sort_data)

        self.enable_repartitioning: Param[bool] = Param(
            self, "enable_repartitioning", ""
        )
        self._setDefault(enable_repartitioning=enable_repartitioning)

        self.operation: Param[Union[FhirSenderOperation, str]] = Param(
            self, "operation", ""
        )
        self._setDefault(operation=operation)

        self.log_level: Param[str] = Param(self, "log_level", "")
        self._setDefault(log_level=log_level)

        self.smart_merge: Param[Optional[bool]] = Param(self, "smart_merge", "")
        self._setDefault(smart_merge=smart_merge)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        transformer: ProxyBase = self.getTransformer()
        func_get_path: Callable[[str, str], str] = self.getFuncGetPath()
        func_get_response_path: Callable[[str, str], str] = (
            self.getFuncGetResponsePath()
        )
        file_format: str = self.getFileFormat()
        fhir_server_url: str = self.getFhirServerUrl()
        fhir_validation_url: Optional[str] = self.getFhirValidationServerUrl()
        name: Optional[str] = self.getName()
        source_entity_name: str = self.getSourceEntityName()
        athena_schema: Optional[str] = self.getAthenaSchema()
        parameters = self.getParameters()
        additional_request_headers: Optional[Dict[str, str]] = (
            self.getAdditionalRequestHeaders()
        )
        sort_data: Optional[Dict[str, Dict[str, Any]]] = self.getOrDefault(
            self.sort_data
        )
        enable_repartitioning: bool = self.getOrDefault(self.enable_repartitioning)
        operation: Union[FhirSenderOperation, str] = self.getOrDefault(self.operation)
        smart_merge: Optional[bool] = self.getSmartMerge()

        assert parameters is not None
        progress_logger = self.getProgressLogger()

        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()
        send_to_fhir: Optional[bool] = self.getSendToFhir()

        mode: str = self.getOrDefault(self.mode)
        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)

        log_level: Optional[str] = self.getOrDefault(self.log_level) or environ.get(
            "LOGLEVEL"
        )

        self.logger.info(f"Calling {fhir_server_url} with client_id={auth_client_id}")

        if len(transformer.transformers) == 0:
            raise AutomapperToFhirTransformerException(
                f"No transformers found for '{transformer.name}'."
            )

        t: Transformer
        for t in transformer.transformers:
            if isinstance(t, FrameworkMappingLoader):
                if hasattr(t, "getName"):
                    # noinspection Mypy
                    stage_name = t.getName()
                    if not stage_name:
                        stage_name = t.__class__.__name__
                else:
                    stage_name = t.__class__.__name__

                await t.transform_async(df)  # run the automapper
                views: List[str] = t.getViews()
                view: str
                for view in views:
                    self.logger.info(
                        f"---- Started processing for view: {view} --------"
                    )
                    # get resource name
                    result_df: DataFrame = df.sparkSession.table(view)
                    # True if sort_data field exists and view name is present as dict key
                    need_sorting: bool = (
                        True if (sort_data and sort_data.get(view)) else False
                    )
                    if spark_is_data_frame_empty(df=result_df):
                        self.logger.info(f"No data to export/send for view '{view}'")
                        continue
                    self.logger.info(f"----- view: {view} has rows-------")
                    first_row = result_df.select("resourceType").first()
                    assert first_row
                    # noinspection PyPep8Naming
                    resourceType: str = first_row["resourceType"]
                    fhir_resource_path: str = func_get_path(view, resourceType)
                    fhir_resource_response_path: str = func_get_response_path(
                        view, resourceType
                    )

                    # export as FHIR to local fhir_resource_path
                    if file_format == "parquet":
                        await FrameworkParquetExporter(
                            name=f"{name}_parquet_exporter",
                            parameters=parameters,
                            progress_logger=progress_logger,
                            view=view,
                            mode=mode,
                            file_path=fhir_resource_path,
                        ).transform_async(df)
                    else:
                        await FhirExporter(
                            view=view,
                            file_path=fhir_resource_path,
                            parameters=parameters,
                            progress_logger=progress_logger,
                            name=f"{name}_fhir_exporter",
                            mode=mode,
                        ).transform_async(df)

                    if fhir_resource_path.startswith("s3") and athena_schema:
                        # create table in Athena
                        s3_temp_folder: str = str(parameters.get("s3_temp_folder"))
                        await AthenaTableCreator(
                            view=view,
                            schema_name=athena_schema,
                            table_name=f"{source_entity_name}_{resourceType}",
                            s3_source_path=fhir_resource_path,
                            source_file_type=AthenaSourceFileType.JSONL,
                            s3_temp_folder=s3_temp_folder,
                            parameters=parameters,
                            progress_logger=progress_logger,
                        ).transform_async(df)
                    # send to FHIR server
                    if send_to_fhir:
                        await FhirSender(
                            resource=resourceType,
                            server_url=fhir_server_url,
                            validation_server_url=fhir_validation_url,
                            file_path=fhir_resource_path,
                            file_format=file_format,
                            response_path=fhir_resource_response_path,
                            parameters=parameters,
                            progress_logger=progress_logger,
                            batch_size=parameters.get("batch_size", 0),
                            throw_exception_on_validation_failure=parameters.get(
                                "throw_exception_on_validation_failure"
                            ),
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            name=f"{name}_fhir_sender",
                            additional_request_headers=additional_request_headers,
                            mode=mode,
                            run_synchronously=run_synchronously,
                            sort_by_column_name_and_type=(
                                sort_data[view].get("sort_by_column_name_and_type")
                                if need_sorting and sort_data
                                else None
                            ),
                            drop_fields_from_json=(
                                sort_data[view].get("drop_fields_from_json")
                                if need_sorting and sort_data
                                else None
                            ),
                            operation=operation,
                            log_level=log_level,
                            smart_merge=smart_merge,
                        ).transform_async(df)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTransformer(self) -> ProxyBase:
        return self.getOrDefault(self.transformer)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetPath(self) -> Callable[[str, str], str]:
        return self.getOrDefault(self.func_get_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFuncGetResponsePath(self) -> Callable[[str, str], str]:
        return self.getOrDefault(self.func_get_response_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFhirServerUrl(self) -> str:
        return self.getOrDefault(self.fhir_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFileFormat(self) -> str:
        return self.getOrDefault(self.file_format)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFhirValidationServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.fhir_validation_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSourceEntityName(self) -> str:
        return self.getOrDefault(self.source_entity_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAthenaSchema(self) -> Optional[str]:
        return self.getOrDefault(self.athena_schema)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSendToFhir(self) -> Optional[bool]:
        return self.getOrDefault(self.send_to_fhir)

    # no inspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalRequestHeaders(self) -> Optional[Dict[str, str]]:
        return self.getOrDefault(self.additional_request_headers)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSmartMerge(self) -> Optional[bool]:
        return self.getOrDefault(self.smart_merge)



from typing import Optional

from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent


class TelemetryParentMixin:
    def __init__(self) -> None:
        self.telemetry_parent: Optional[TelemetryParent] = None

    def set_telemetry_parent(
        self, *, telemetry_parent: Optional[TelemetryParent]
    ) -> None:
        self.telemetry_parent = telemetry_parent



from typing import List, Type

import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.geocodio_standardizing_vendor_api_response import (
    GeocodioStandardizingVendorApiResponse,
)

logger = structlog.get_logger(__file__)


class GeocodioStandardizingVendor(
    StandardizingVendor[GeocodioStandardizingVendorApiResponse]
):
    @classmethod
    def get_api_response_class(cls) -> Type[GeocodioStandardizingVendorApiResponse]:
        return GeocodioStandardizingVendorApiResponse

    async def standardize_async(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse[GeocodioStandardizingVendorApiResponse]]:
        vendor_specific_addresses: List[GeocodioStandardizingVendorApiResponse] = [
            GeocodioStandardizingVendorApiResponse.from_raw_address(raw_address)
            for raw_address in raw_addresses
        ]

        vendor_responses: List[
            VendorResponse[GeocodioStandardizingVendorApiResponse]
        ] = self._to_vendor_response(
            vendor_response=vendor_specific_addresses,
            raw_addresses=raw_addresses,
            vendor_name=self.get_vendor_name(),
            response_version=self.get_version(),
        )
        return vendor_responses

    def vendor_specific_to_std(
        self,
        *,
        vendor_specific_addresses: List[
            VendorResponse[GeocodioStandardizingVendorApiResponse]
        ],
        raw_addresses: List[RawAddress],
    ) -> List[StandardizedAddress]:
        """
        each vendor class knows how to convert its response to StdAddress
        """
        std_addresses = [
            StandardizedAddress(
                address_id=a.api_call_response.RecordID,
                line1=a.api_call_response.line1,
                line2=a.api_call_response.line2,
                city=a.api_call_response.city,
                zipcode=a.api_call_response.zipcode,
                state=a.api_call_response.state,
                country=a.api_call_response.geocodio_Country,
                latitude=(
                    str(a.api_call_response.Latitude)
                    if a.api_call_response.Latitude
                    else None
                ),
                longitude=(
                    str(a.api_call_response.Longitude)
                    if a.api_call_response.Longitude
                    else None
                ),
                standardize_vendor="Geocodio",
                county=None,
                formatted_address=None,
            )
            for a in vendor_specific_addresses
        ]
        return std_addresses

    @classmethod
    def get_vendor_name(cls) -> str:
        return "Geocodio"

    def batch_request_max_size(self) -> int:
        return 100

    def _to_vendor_response(
        self,
        vendor_response: List[GeocodioStandardizingVendorApiResponse],
        raw_addresses: List[RawAddress],
        vendor_name: str,
        response_version: str,
    ) -> List[VendorResponse[GeocodioStandardizingVendorApiResponse]]:
        # create the map
        id_response_map = {a.get_id(): a for a in raw_addresses}
        # find and assign
        return [
            VendorResponse(
                api_call_response=r,
                related_raw_address=id_response_map[r.RecordID],
                vendor_name=vendor_name,
                response_version=response_version,
            )
            for r in vendor_response
        ]



import json
from datetime import datetime, timedelta, UTC
from logging import getLogger
from typing import Any, Optional

# noinspection PyUnresolvedReferences
from slack_sdk.web.client import WebClient

# noinspection PyUnresolvedReferences
from slack_sdk.web.slack_response import SlackResponse

# noinspection PyUnresolvedReferences
from slack_sdk.errors import SlackApiError

from spark_pipeline_framework.utilities.slack.base_slack_client import BaseSlackClient


logger = getLogger(__name__)


class SlackClientNative(BaseSlackClient):
    """
    Implements a native Slack client
    """

    def __init__(self, slack_token: str, channel: str, bot_user_name: str) -> None:
        """
        Client for sending messages to Slack

        :param slack_token: token to auth with Slack
        :param channel: channel to post message into
        :param bot_user_name: user name to use when posting messages
        """
        self.slack_token: str = slack_token
        self.slack_channel: str = channel
        self.slack_icon_url: str = "https://www.flaticon.com/free-icon/freepik_23317"
        self.slack_user_name: str = bot_user_name
        self.slack_thread: Optional[str] = None
        # if we have to wait till a time before slack will let us send messages
        self.wait_till_datetime: Optional[datetime] = None

    def post_message_to_slack(
        self,
        text: str,
        blocks: Any = None,
        use_conversation_threads: bool = True,
        slack_thread: Optional[str] = None,
    ) -> Any:
        """
        Posts a message to Slack
        :param text: message to post.  Can be markdown.
        :param blocks: (Optional) blocks to post
        :param use_conversation_threads: whether to send messages as reply to first message
        :return: response from Slack
        """
        if not text:
            return None

        if (
            self.wait_till_datetime is not None
            and self.wait_till_datetime > datetime.now(UTC)
        ):
            return None

        try:
            web = WebClient(token=self.slack_token)
            response: SlackResponse = web.chat_postMessage(
                text=text,
                channel=self.slack_channel,
                thread_ts=(
                    slack_thread if slack_thread is not None else self.slack_thread
                ),
                icon_url=self.slack_icon_url,
                username=self.slack_user_name,
                blocks=json.dumps(blocks) if blocks else None,
            )

            if not self.slack_thread and use_conversation_threads:
                if (
                    response.status_code == 200
                    and response.data
                    and "ts" in response.data
                ):
                    self.slack_thread = response.data["ts"]  # type: ignore
            self.wait_till_datetime = None  # clear this on success
            return response
        except SlackApiError as e:
            logger.warning(f"Slack API Error: {e.response['error']}")
            if e.response.status_code == 429:
                # The `Retry-After` header will tell you how long to wait before retrying
                delay_in_seconds: int = int(e.response.headers["Retry-After"])
                logger.warning(f"Rate limited. Retrying in {delay_in_seconds} seconds")
                self.wait_till_datetime = datetime.now(UTC) + timedelta(
                    seconds=delay_in_seconds
                )
            return None
        except Exception as e:
            logger.warning(f"Unknown error calling Slack API: {e}")
            return None



from typing import Iterable, List, Any, Optional, cast

from pyspark.sql import DataFrame


def get_pretty_table(
    iterable: Iterable[List[Any]], header: Iterable[str], name: Optional[str] = None
) -> str:
    """
    Pretty prints an iterable with headers


    :param iterable: iterable to print
    :param header: headers to use
    :param name:
    :return:
    """
    max_len: List[int] = [len(x) for x in header]
    output: str = ""
    if name:
        output += "-" * (sum(max_len) + 1) + "\n"
        output += name + "\n"
    row: List[Any]
    # first write out the correct number of - at the top
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        for index, col in enumerate(row):
            if max_len[index] < len(str(col)):
                max_len[index] = len(str(col))
    output += "-" * (sum(max_len) + 1) + "\n"
    # Then write out the headers
    output += (
        "|"
        + "".join([h + " " * (l - len(h)) + "|" for h, l in zip(header, max_len)])
        + "\n"
    )
    # Then write the dashes under the headers
    output += "-" * (sum(max_len) + 1) + "\n"
    # Then write out the contents of each row
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        output += (
            "|"
            + "".join(
                [str(c) + " " * (l - len(str(c))) + "|" for c, l in zip(row, max_len)]
            )
            + "\n"
        )
    # Finally write out the - at the bottom
    output += "-" * (sum(max_len) + 1) + "\n"
    return output


def get_pretty_data_frame_old(
    df: DataFrame, limit: int, name: Optional[str] = None
) -> str:
    """
    Returns the dataframe as a string

    :param df:
    :param limit:
    :param name: name to show
    :return:
    """
    if limit == 0:
        return ""

    rows: List[List[Any]] = [list(row) for row in df.limit(limit).collect()]

    return get_pretty_table(rows, df.columns, name=name)


def get_show_string(
    df: DataFrame, n: int = 20, truncate: bool = True, vertical: bool = False
) -> str:
    """
    Returns dataframe as a string
    Based on https://spark.apache.org/docs/latest/api/python/_modules/pyspark/sql/dataframe.html#DataFrame.show
    """
    if isinstance(truncate, bool) and truncate:
        # noinspection PyProtectedMember
        return cast(str, df._jdf.showString(n, 20, vertical))
    else:
        # noinspection PyProtectedMember
        return cast(str, df._jdf.showString(n, int(truncate), vertical))


def get_pretty_data_frame(df: DataFrame, limit: int, name: Optional[str] = None) -> str:
    """
    Returns the dataframe as a string

    :param df:
    :param limit:
    :param name: name to show
    :return:
    """
    if limit == 0:
        return ""

    output: str = ""
    if name:
        output += "-" * 100 + "\n"
        output += name + "\n"

    output += get_show_string(df, n=limit, truncate=False)
    return output


def get_data_frame_as_csv(df: DataFrame, limit: int) -> str:
    """
    Returns the dataframe as a string

    :param df:
    :param limit:
    :return:
    """
    if limit == 0:
        return ""

    rows: List[List[Any]] = [list(row) for row in df.limit(limit).collect()]

    result: str = ""
    result += ",".join(['"' + c + '"' for c in df.columns]) + "\n"
    row: List[Any]
    for row in rows:
        result += ",".join(['"' + str(r) + '"' for r in row]) + "\n"
    return result



import json
from logging import Logger
from typing import Any, Dict, Iterator, List, Optional, Union, cast

from opensearchpy import AsyncOpenSearch, ConnectionTimeout
from furl import furl
from opensearchpy.helpers import async_bulk
from opensearchpy.helpers.errors import BulkIndexError

from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_result import (
    ElasticSearchResult,
)
from spark_pipeline_framework.utilities.elastic_search.elastic_search_connection import (
    ElasticSearchConnection,
)


class ElasticSearchHelpers:
    @staticmethod
    def generate_es_bulk(
        *,
        json_data_list: List[str],
        operation: str,
        index: str,
        doc_id_prefix: Optional[str] = None,
    ) -> Iterator[Dict[str, Any]]:
        """this is how ES bulk function likes the input be served
        https://elasticsearch-py.readthedocs.io/en/v7.12.0/helpers.html#bulk-helpers
        """
        for item in json_data_list:
            item_json = json.loads(item)
            try:
                doc_id = item_json["id"]
                if doc_id_prefix:
                    doc_id = f"{doc_id_prefix}-{doc_id}"

                if operation != "index":
                    wrapped_doc: Dict[str, Any] = {
                        "_op_type": operation,
                        "_index": index,
                        "_id": doc_id,
                        "doc": item_json,
                        "doc_as_upsert": True,
                    }
                    item_json = wrapped_doc

                if "_id" not in item_json.keys():
                    item_json["_id"] = doc_id
            except KeyError:
                # just for local logging not usable for cluster env
                print(f"item without id >>> {item}")
                # raise e

            yield item_json

    @staticmethod
    async def send_json_bundle_to_elasticsearch_async(
        *,
        json_data_list: List[str],
        index: str,
        operation: str,
        logger: Logger,
        doc_id_prefix: Optional[str] = None,
        timeout: int = 60,
    ) -> ElasticSearchResult:
        """
        Sends a list of json strings to the server
        (Runs on worker nodes so any logging will not show)

        :param json_data_list: list of json strings to send to server
        :param index: name of ElasticSearch index
        :param operation:
        :param logger: logger
        :param doc_id_prefix: a string to be prepended to the _id field for a document
        :param timeout: timeout in seconds for calls made by this client
        """
        assert isinstance(json_data_list, list)
        assert all(isinstance(j, str) for j in json_data_list)
        assert index is not None
        assert isinstance(index, str)
        assert operation is not None
        assert isinstance(operation, str)
        assert doc_id_prefix is None or isinstance(doc_id_prefix, str)

        es_connection = ElasticSearchConnection()
        es_client: AsyncOpenSearch = es_connection.get_elastic_search_async_client(
            timeout=timeout
        )
        server_url = es_connection.get_elastic_search_host()

        payload: List[Dict[str, Any]] = list(
            ElasticSearchHelpers.generate_es_bulk(
                json_data_list=json_data_list,
                operation=operation,
                index=index,
                doc_id_prefix=doc_id_prefix,
            )
        )
        success: int = 0
        failed: Union[int, List[Any]] = 0
        errors: List[str] = []

        try:
            # https://github.com/opensearch-project/opensearch-py/blob/main/opensearchpy/_async/helpers/actions.py
            success, failed = await async_bulk(
                client=es_client,
                actions=payload,
                index=index,
                stats_only=True,
                request_timeout=60,
            )
            assert isinstance(success, int)
            # since we passed stats_only=True, failed will be an int
            assert isinstance(failed, int)
        except BulkIndexError as e:
            for error in e.errors:
                logger.error(f"The following record failed to index: {error}")
                errors.append(error)
        except ConnectionTimeout as e:
            logger.error(f"ConnectionTimeout: {e}")
            failed = len(payload)
        finally:
            # close connection manually to avoid this error:
            # https://elasticsearch-py.readthedocs.io/en/v7.14.0/async.html#receiving-unclosed-client-session-connector-warning
            await es_client.close()

        full_uri: furl = furl(server_url)
        full_uri /= index
        return ElasticSearchResult(
            url=full_uri.url,
            success=success,
            failed=cast(int, failed),
            payload=list(payload),
            partition_index=0,
            error=json.dumps(errors) if len(errors) > 0 else None,
        )



from pathlib import Path
from typing import Dict, Any, Optional, Union

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.errors import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FhirTextReader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str],
        destination_column_name: str = "value",
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        bad_records_path: Optional[Union[Path, str]] = None,
        create_empty_view_if_file_path_not_found: Optional[bool] = False,
        limit: int = -1,
    ):
        """
        Reads files from path and creates one column called destination_column_name that contains the FHIR data as text


        :param file_path: path to the fhir data to read
        :param view: the view to load the data into
        :param name: a name to use when logging information about this transformer
        :param destination_column_name: column to create
        :param parameters: the parameter dictionary
        :param progress_logger: the progress logger
        :param bad_records_path: the path to write files that spark cannot parse
        :param create_empty_view_if_file_path_not_found: create an empty view using the provided schema in cases where the
        file_path may not exist. *NOTE if this is true and schema is not provided then the df will have no columns!
        """

        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert isinstance(file_path, Path) or isinstance(file_path, str)

        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[Union[Path, str]] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.destination_column_name: Param[str] = Param(
            self, "destination_column_name", ""
        )
        self._setDefault(destination_column_name=destination_column_name)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.bad_records_path: Param[Optional[Union[Path, str]]] = Param(
            self, "bad_records_path", ""
        )
        self._setDefault(bad_records_path=None)

        self.create_empty_view_if_file_path_not_found: Param[Optional[bool]] = Param(
            self, "create_empty_view_if_file_path_not_found", ""
        )
        self._setDefault(
            create_empty_view_if_file_path_not_found=create_empty_view_if_file_path_not_found
        )

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        path: Union[Path, str] = self.getFilePath()
        name: Optional[str] = self.getName()
        destination_column_name: str = self.getDestinationColumnName()
        bad_records_path: Optional[Union[Path, str]] = self.getBadRecordsPath()
        create_empty_view_if_file_path_not_found: Optional[bool] = (
            self.getCreateEmptyViewIfFilePathNotFound()
        )

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        # limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_fhir_text_reader", progress_logger=progress_logger
        ):
            try:
                # drop malformed records:
                # https://spark.apache.org/docs/2.3.1/api/java/org/apache/spark/sql/DataFrameReader.html#json-scala.collection.Seq-
                # mode (default PERMISSIVE): allows a mode for dealing with corrupt records during parsing.
                # PERMISSIVE : when it meets a corrupted record, puts the malformed string into a field configured
                #   by columnNameOfCorruptRecord, and sets other fields to null. To keep corrupt records,
                #   an user can set a string type field named columnNameOfCorruptRecord in an user-defined schema.
                #   If a schema does not have the field, it drops corrupt records during parsing.
                #   When inferring a schema, it implicitly adds a columnNameOfCorruptRecord field in an
                #   output schema.
                # DROPMALFORMED : ignores the whole corrupted records.
                # FAILFAST : throws an exception when it meets corrupted records.
                # https://docs.databricks.com/spark/latest/spark-sql/handling-bad-records.html
                reader = df.sparkSession.read
                if bad_records_path:
                    reader = reader.option("badRecordsPath", str(bad_records_path))
                df = reader.text(str(path))

                if destination_column_name != "value":
                    df = df.withColumnRenamed("value", destination_column_name)

                if view:
                    df.createOrReplaceTempView(view)

                self.logger.info(f"Read {df.count()} records from text file {path}")

            except AnalysisException as e:
                self.logger.exception(f"File read failed from {path}")
                if create_empty_view_if_file_path_not_found and view:
                    df.createOrReplaceTempView(view)
                else:
                    raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDestinationColumnName(self) -> str:
        return self.getOrDefault(self.destination_column_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[Path, str]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBadRecordsPath(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.bad_records_path)

    # noinspection PyPep8Naming
    def getCreateEmptyViewIfFilePathNotFound(self) -> Optional[bool]:
        return self.getOrDefault(self.create_empty_view_if_file_path_not_found)



from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_parameters import (
    FhirReceiverParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)


def get_fhir_receiver_parameters() -> FhirReceiverParameters:
    return FhirReceiverParameters(
        total_partitions=1,
        batch_size=10,
        has_token_col=False,
        server_url="http://fhir-server",
        log_level="DEBUG",
        action=None,
        action_payload=None,
        additional_parameters=None,
        filter_by_resource=None,
        filter_parameter=None,
        sort_fields=None,
        auth_server_url="http://fhir-server/token",
        auth_client_id=None,
        auth_client_secret=None,
        auth_login_token=None,
        auth_scopes=None,
        auth_well_known_url=None,
        include_only_properties=None,
        separate_bundle_resources=False,
        expand_fhir_bundle=False,
        accept_type=None,
        content_type=None,
        additional_request_headers=None,
        accept_encoding=None,
        slug_column=None,
        retry_count=None,
        exclude_status_codes_from_retry=None,
        limit=None,
        auth_access_token=None,
        resource_type="Patient",
        error_view=None,
        url_column=None,
        use_data_streaming=None,
        graph_json=None,
        ignore_status_codes=[],
        refresh_token_function=None,
        use_id_above_for_paging=True,
        pandas_udf_parameters=AsyncPandasUdfParameters(),
    )



from typing import Dict, Any, Optional

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import (
    struct,
    lit,
    col,
    concat,
    array,
    regexp_replace,
    substring,
)
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkFhirMetaUpdater(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        resource_type: str,
        source_view: str,
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        slug_column: str,
        url_column: str,
        connection_type: str,
    ):
        """
        Used to add/update the meta field with the b.well mandatory stuff


        :param resource_type: type of resource to update
        :param source_view: read from this view
        :param view: write to this view
        :param slug_column: use this column to set the security tags
        :param url_column: column to read the url
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.resource_type: Param[str] = Param(self, "resource_type", "")
        self._setDefault(resource_type=None)

        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=None)

        self.slug_column: Param[str] = Param(self, "slug_column", "")
        self._setDefault(slug_column=None)

        self.url_column: Param[str] = Param(self, "url_column", "")
        self._setDefault(url_column=None)

        self.connection_type: Param[str] = Param(self, "connection_type", "")
        self._setDefault(connection_type=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        source_view: str = self.getSourceView()
        resource_type: str = self.getResourceType()
        slug_column: str = self.getOrDefault(self.slug_column)
        url_column: str = self.getOrDefault(self.url_column)
        connection_type: str = self.getOrDefault(self.connection_type)

        df = df.sparkSession.table(source_view)

        # The schema will be:
        # { "patient":{"id":1}, "url":"https://foo", "service_slug":"aetna"}

        df = df.withColumn(
            "id",
            substring(
                regexp_replace(
                    string=concat(col(slug_column), lit("_"), col("id")),
                    pattern=r"[^A-Za-z0-9\-\.]",
                    replacement="-",
                ),
                0,
                63,
            ),
        )
        owner_codeset: str = "https://www.icanbwell.com/owner"
        access_codeset: str = "https://www.icanbwell.com/access"
        vendor_codeset: str = "https://www.icanbwell.com/vendor"
        connection_type_codeset: str = "https://www.icanbwell.com/connectionType"

        df = df.withColumn(
            "meta",
            struct(
                concat(col(url_column), lit(f"/{resource_type}/"), col("id")).alias(
                    "source"
                ),
                array(
                    struct(
                        lit(owner_codeset).alias("system"),
                        lit(col(slug_column)).alias("code"),
                    ),
                    struct(
                        lit(access_codeset).alias("system"),
                        lit(col(slug_column)).alias("code"),
                    ),
                    struct(
                        lit(vendor_codeset).alias("system"),
                        lit(col(slug_column)).alias("code"),
                    ),
                    struct(
                        lit(connection_type_codeset).alias("system"),
                        lit(f"{connection_type}").alias("code"),
                    ),
                ).alias("security"),
            ),
        )

        # drop the extra columns
        df = df.drop(url_column, slug_column)
        df.createOrReplaceTempView(view)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSourceView(self) -> str:
        return self.getOrDefault(self.source_view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResourceType(self) -> str:
        return self.getOrDefault(self.resource_type)



import json
import os
from typing import Any, Dict, List, Union, Optional, Mapping

from helixtelemetry.telemetry.context.telemetry_context import TelemetryContext
from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.providers.null_telemetry import NullTelemetry
from helixtelemetry.telemetry.providers.open_telemetry import OpenTelemetry
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_attribute_value import (
    TelemetryAttributeValue,
)
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent
from pyspark.ml.base import Transformer
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.log_level import LogLevel
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.class_helpers import ClassHelpers
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


class FrameworkPipeline(Transformer, LoopIdMixin, TelemetryParentMixin):
    def __init__(
        self,
        parameters: Dict[str, Any],
        progress_logger: ProgressLogger,
        run_id: Optional[str] = None,
        log_level: Optional[Union[int, str]] = None,
        telemetry_enable: Optional[bool] = None,
        telemetry_context: Optional[TelemetryContext] = None,
        name: Optional[str] = None,
        attributes: Optional[Mapping[str, TelemetryAttributeValue]] = None,
        telemetry_parent: Optional[TelemetryParent] = None,
    ) -> None:
        """
        Base class for all pipelines
        :param parameters:
        :param progress_logger:
        """
        super(FrameworkPipeline, self).__init__()
        self.transformers: List[Transformer] = []
        self._run_id: Optional[str] = run_id
        self.steps: List[Union[Transformer, List[Transformer]]] = []
        self.__parameters: Dict[str, Any] = parameters
        self.progress_logger: ProgressLogger = progress_logger
        self.loop_id: Optional[str] = None
        self.log_level: Optional[Union[int, str]] = log_level or os.environ.get(
            "LOGLEVEL"
        )
        self.telemetry_enable: Optional[bool] = telemetry_enable or bool(
            os.environ.get("TELEMETRY_ENABLE")
        )

        if telemetry_parent:
            self.telemetry_parent = telemetry_parent
        else:
            self.set_telemetry_parent(
                telemetry_parent=TelemetryParent(
                    name=name or self.__class__.__qualname__,
                    trace_id=None,
                    span_id=None,
                    telemetry_context=(
                        telemetry_context
                        or TelemetryContext(
                            provider=(
                                OpenTelemetry.telemetry_provider
                                if self.telemetry_enable
                                else NullTelemetry.telemetry_provider
                            ),
                            service_name=os.getenv(
                                "OTEL_SERVICE_NAME", "helix-pipelines"
                            ),
                            environment=os.getenv("ENV", "development"),
                            attributes=attributes,
                            log_level=log_level,
                            instance_name=os.getenv(
                                "OTEL_INSTANCE_NAME",
                                self.parameters.get("flow_run_name", "unknown"),
                            ),
                            service_namespace=os.getenv(
                                "OTEL_SERVICE_NAMESPACE", "helix-pipelines"
                            ),
                        )
                    ),
                    attributes=attributes,
                )
            )

        self.name: Optional[str] = name
        self.attributes: Mapping[str, TelemetryAttributeValue] = attributes or {}
        self.attributes = {k: v for k, v in self.attributes.items()} | {
            "run_id": self._run_id
        }

    @property
    def parameters(self) -> Dict[str, Any]:
        return self.__parameters

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame) -> "FrameworkPipeline":
        return self

    def _transform(self, df: DataFrame) -> DataFrame:
        """
        Override this method to implement transformation

        :param df: input dataframe
        :return: transformed dataframe
        """
        return AsyncHelper.run(self._transform_async(df))

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        """
        Runs all the transformers in the pipeline in sequence on the input DataFrame and returns the transformed DataFrame


        """
        telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
            telemetry_parent=self.telemetry_parent or TelemetryParent.get_null_parent()
        ).create_telemetry_span_creator(log_level=self.log_level)

        telemetry_span: TelemetrySpanWrapper
        async with telemetry_span_creator.create_telemetry_span_async(
            name=self.name or self.__class__.__qualname__,
            attributes=self.attributes,
            telemetry_parent=self.telemetry_parent,
        ) as telemetry_span:
            # if steps are defined but not transformers then convert steps to transformers first
            if len(self.steps) > 0 and len(self.transformers) == 0:
                self.transformers = self.create_steps(self.steps)
            # get the logger to use
            logger = get_logger(__name__)
            count_of_transformers: int = len(self.transformers)
            i: int = 0
            pipeline_name: str = self.__class__.__name__
            self.progress_logger.log_event(
                event_name=pipeline_name,
                event_text=(
                    f"Starting Pipeline {pipeline_name}" + f"_{self._run_id}"
                    if self._run_id
                    else ""
                ),
                log_level=LogLevel.INFO,
            )
            for transformer in self.transformers:
                assert isinstance(transformer, Transformer), type(transformer)
                if hasattr(transformer, "getName"):
                    # noinspection Mypy
                    stage_name = (
                        f"{transformer.getName()} ({transformer.__class__.__name__})"
                    )
                else:
                    stage_name = transformer.__class__.__name__

                transformer_span: TelemetrySpanWrapper
                async with telemetry_span_creator.create_telemetry_span_async(
                    name=stage_name,
                    attributes={
                        "loop_id": self.loop_id,
                    },
                    telemetry_parent=telemetry_span.create_child_telemetry_parent(),
                ) as transformer_span:
                    try:
                        i += 1
                        logger.info(
                            f"---- Running pipeline [{pipeline_name}] transformer [{stage_name}]  "
                            f"({i} of {count_of_transformers}) ----"
                        )
                        if isinstance(transformer, LoopIdMixin):
                            transformer.set_loop_id(self.loop_id)
                        if isinstance(transformer, TelemetryParentMixin):
                            transformer.set_telemetry_parent(
                                telemetry_parent=transformer_span.create_child_telemetry_parent()
                            )

                        with ProgressLogMetric(
                            progress_logger=self.progress_logger,
                            name=str(stage_name) or "unknown",
                        ):
                            self.progress_logger.log_event(
                                pipeline_name,
                                event_text=f"Running pipeline step {stage_name}",
                            )
                            if hasattr(transformer, "_transform_async"):
                                # noinspection PyProtectedMember
                                df = await transformer._transform_async(df=df)
                            else:
                                df = transformer.transform(dataset=df)

                    except Exception as e:
                        logger.error(
                            f"!!!!!!!!!!!!! pipeline [{pipeline_name}] transformer [{stage_name}] threw exception !!!!!!!!!!!!!"
                        )
                        # use exception chaining to add stage name but keep original exception
                        # friendly_spark_exception: FriendlySparkException = (
                        #     FriendlySparkException(exception=e, stage_name=stage_name)
                        # )
                        # error_messages: List[str] = (
                        #     friendly_spark_exception.message.split("\n")
                        #     if friendly_spark_exception.message
                        #     else []
                        # )
                        # for error_message in error_messages:
                        #     logger.error(msg=error_message)

                        if hasattr(transformer, "getSql"):
                            # noinspection Mypy
                            logger.error(transformer.getSql())
                        logger.error(
                            "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
                        )
                        self.progress_logger.log_exception(
                            event_name=pipeline_name,
                            event_text=f"Exception in Stage={stage_name}",
                            ex=e,
                        )
                        # if hasattr(e, "message"):
                        #     e.message = f"Exception in stage {stage_name}" + e.message
                        if len(e.args) >= 1:
                            # e.args = (e.args[0] + f" in stage {stage_name}") + e.args[1:]
                            e.args = (f"In Stage ({stage_name})", *e.args)
                        raise e

            self.progress_logger.log_event(
                event_name=pipeline_name,
                event_text=f"Finished Pipeline {pipeline_name}",
                log_level=LogLevel.INFO,
            )

            await telemetry_span_creator.flush_async()
            return df

    # noinspection PyMethodMayBeStatic
    def create_steps(
        self,
        my_list: Union[
            List[Transformer],
            List[FrameworkTransformer],
            List[Union[Transformer, List[Transformer]]],
            List[Union[FrameworkTransformer, List[FrameworkTransformer]]],
            # List[DefaultParamsReadable[Any]],
        ],
    ) -> List[Transformer]:
        return create_steps(my_list)

    def finalize(self) -> None:
        pass

    def as_dict(self) -> Dict[str, Any]:
        return {
            "short_type": self.__class__.__name__,
            "type": ClassHelpers.get_full_name_of_instance(self),
            # self.parameters is a subclass of dict so json.dumps thinks it can't serialize it
            "params": {
                k: v if not hasattr(v, "as_dict") else v.as_dict()
                for k, v in self.parameters.items()
            },
            "steps": [
                s.as_dict() if not isinstance(s, list) else [s1.as_dict() for s1 in s]  # type: ignore
                for s in self.steps
            ],
        }

    def __str__(self) -> str:
        return json.dumps(self.as_dict(), default=str)

    async def transform_async(self, df: DataFrame) -> DataFrame:
        return await self._transform_async(df=df)



import dataclasses
from typing import TypeVar, Generic, Optional

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)

T = TypeVar("T")


@dataclasses.dataclass
class VendorResponse(Generic[T]):
    """
    This class is used to store the response from the vendor.  It is a generic class where you can
    specify the type of the class that stores the api_call_response

    """

    vendor_name: str
    """name of the vendor"""
    response_version: str
    """version of the response"""
    api_call_response: T
    """response from the vendor"""
    related_raw_address: RawAddress
    """related raw address"""
    error: Optional[str] = None
    """error message if any"""



from os import environ
from typing import Dict, Any, Optional, Union, List, Callable

from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent

from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkTransformerGroup(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        enable: Union[bool, Callable[[DataFrame], bool]] = True,
        enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = None,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Used to group transformers

        :param stages: list of transformers or a function that returns a list of transformers
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.enable: Union[bool, Callable[[DataFrame], bool]] = enable

        self.enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = enable_if_view_not_empty

        self.logger = get_logger(__name__)

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        enable = self.enable(df) if callable(self.enable) else self.enable
        view_enable_if_view_not_empty = (
            self.enable_if_view_not_empty(self.loop_id)
            if callable(self.enable_if_view_not_empty)
            else self.enable_if_view_not_empty
        )
        enable_if_view_not_empty = (
            (
                df.sparkSession.catalog.tableExists(view_enable_if_view_not_empty)
                and not df.sparkSession.table(view_enable_if_view_not_empty).isEmpty()
            )
            if view_enable_if_view_not_empty
            else True
        )

        telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
            telemetry_parent=self.telemetry_parent or TelemetryParent.get_null_parent()
        ).create_telemetry_span_creator(log_level=environ.get("LOGLEVEL"))

        if (enable or enable is None) and enable_if_view_not_empty:
            stages: List[Transformer] = (
                self.stages if not callable(self.stages) else self.stages()
            )
            for stage in stages:
                if hasattr(stage, "getName"):
                    # noinspection Mypy
                    stage_name = stage.getName()
                else:
                    stage_name = stage.__class__.__name__

                telemetry_span: TelemetrySpanWrapper
                async with telemetry_span_creator.create_telemetry_span_async(
                    name=stage_name,
                    attributes={},
                    telemetry_parent=self.telemetry_parent,
                ) as telemetry_span:
                    if isinstance(stage, LoopIdMixin):
                        stage.set_loop_id(self.loop_id)
                    if isinstance(stage, TelemetryParentMixin):
                        stage.set_telemetry_parent(
                            telemetry_parent=telemetry_span.create_child_telemetry_parent()
                        )

                    try:
                        if hasattr(stage, "transform_async"):
                            await stage.transform_async(df)
                        else:
                            df = stage.transform(df)
                    except Exception as e:
                        if len(e.args) >= 1:
                            # e.args = (e.args[0] + f" in stage {stage_name}") + e.args[1:]
                            e.args = (f"In Stage ({stage_name})", *e.args)
                        raise e
        else:
            if progress_logger is not None:
                progress_logger.write_to_log(
                    self.getName() or "FrameworkTransformerGroup",
                    f"Skipping stages because enable {self.enable or self.enable_if_view_not_empty} did not evaluate to True",
                )
        return df

    def as_dict(self) -> Dict[str, Any]:
        return {
            **(super().as_dict()),
            "enable": self.enable,
            "enable_if_view_not_empty": self.enable_if_view_not_empty,
            "stages": (
                [s.as_dict() for s in self.stages]  # type: ignore
                if not callable(self.stages)
                else str(self.stages)
            ),
        }



from importlib import import_module
from typing import Any


def load_module(module_path: str, module_name: str) -> Any:
    try:
        module: Any = import_module(name=module_path)
        return getattr(module, module_name)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(f"A Module for {module_name} is not found!")



from __future__ import annotations

import json
from typing import (
    Iterator,
    Any,
    Dict,
    List,
    Callable,
    Iterable,
    cast,
    Optional,
)

import pandas as pd

from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_base_pandas_udf import (
    AsyncBasePandasUDF,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.function_types import (
    HandlePandasDataFrameBatchFunction,
    AcceptedParametersType,
)


class AsyncPandasDataFrameUDF[TParameters: AcceptedParametersType](
    AsyncBasePandasUDF[
        TParameters, pd.DataFrame, pd.DataFrame, Dict[str, Any], Dict[str, Any]
    ]
):

    def __init__(
        self,
        *,
        async_func: HandlePandasDataFrameBatchFunction[TParameters],
        parameters: Optional[TParameters],
        pandas_udf_parameters: AsyncPandasUdfParameters,
    ) -> None:
        super().__init__(
            async_func=async_func,
            parameters=parameters,
            pandas_udf_parameters=pandas_udf_parameters,
        )

    async def get_input_values_from_chunk(
        self, batch: pd.DataFrame
    ) -> List[Dict[str, Any]]:
        pdf_json: str = batch.to_json(orient="records")
        input_values: List[Dict[str, Any]] = json.loads(pdf_json)
        return input_values

    async def create_output_from_dict(
        self, output_values: List[Dict[str, Any]]
    ) -> pd.DataFrame:
        return pd.DataFrame(output_values)

    def get_pandas_udf(
        self,
    ) -> Callable[[Iterable[pd.DataFrame]], Iterator[pd.DataFrame]]:
        """
        Returns a Pandas UDF function that can be used in Spark.

        :return: a Pandas UDF function
        """
        return cast(
            Callable[[Iterable[pd.DataFrame]], Iterator[pd.DataFrame]],
            self.apply_process_partition_udf,
        )



import json
from os import environ, path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import Any, List, Dict, Optional

import pytest

from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.utilities.fhir_helper import FhirHelper
from helix_fhir_client_sdk.utilities.practitioner_generator import PractitionerGenerator
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, ArrayType, StringType
from spark_fhir_schemas.r4.resources.endpoint import EndpointSchema
from spark_fhir_schemas.r4.resources.group import GroupSchema
from spark_fhir_schemas.r4.resources.insuranceplan import InsurancePlanSchema
from spark_fhir_schemas.r4.resources.location import LocationSchema
from spark_fhir_schemas.r4.resources.measurereport import MeasureReportSchema
from spark_fhir_schemas.r4.resources.organization import OrganizationSchema
from spark_fhir_schemas.r4.resources.practitioner import PractitionerSchema
from spark_fhir_schemas.r4.resources.practitionerrole import PractitionerRoleSchema
from spark_fhir_schemas.r4.resources.schedule import ScheduleSchema

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_reader.v1.fhir_reader import FhirReader
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver import (
    FhirReceiver,
)
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


@pytest.mark.parametrize("run_synchronously", [True, False])
@pytest.mark.parametrize("use_data_streaming", [True, False])
async def test_async_real_fhir_server_get_graph_by_id_large(
    spark_session: SparkSession, run_synchronously: bool, use_data_streaming: bool
) -> None:
    print()
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("../temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    patient_json_path: Path = temp_folder.joinpath("patient.json")
    resource_type = "Practitioner"
    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:

        environ["LOGLEVEL"] = "DEBUG"

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url

        fhir_client = await fhir_server_test_context.create_fhir_client_async()
        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        count: int = 100
        roles_per_practitioner: int = 10

        id_dict: Dict[str, List[str]] = PractitionerGenerator.get_ids(
            count=count, roles_per_practitioner=roles_per_practitioner
        )
        # delete Practitioner resources
        print(f"Deleting {count} {resource_type} resources: {id_dict[resource_type]}")
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type=resource_type,
            id_list=id_dict[resource_type],
        )
        # delete Organization resources
        print(f"Deleting {count} Organization resources: {id_dict['Organization']}")
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type="Organization",
            id_list=id_dict["Organization"],
        )
        # delete PractitionerRole resources
        print(
            f"Deleting {count} PractitionerRole resources: {id_dict['PractitionerRole']}"
        )
        await FhirHelper.delete_resources_by_ids_async(
            fhir_client=fhir_client,
            resource_type="PractitionerRole",
            id_list=id_dict["PractitionerRole"],
        )
        print("Finished deleting resources")

        fhir_client = await fhir_server_test_context.create_fhir_client_async()
        fhir_client = fhir_client.url(fhir_server_url).resource(resource_type)

        bundle: Dict[str, Any] = PractitionerGenerator.generate_resources_bundle(
            count=count, roles_per_practitioner=roles_per_practitioner
        )

        expected_resource_count = len(bundle["entry"])
        print("expected_resource_count", expected_resource_count)

        print(f"Merging bundle with {expected_resource_count} resources")
        merge_response: Optional[FhirMergeResponse] = (
            await FhirMergeResponse.from_async_generator(
                fhir_client.merge_async(json_data_list=[json.dumps(bundle)])
            )
        )
        assert merge_response is not None
        print(f"Merged {expected_resource_count} resources")
        print(json.dumps(merge_response.responses))
        assert merge_response.status == 200, merge_response.responses
        assert (
            len(merge_response.responses) == expected_resource_count
        ), merge_response.responses

        # assert merge_response.responses[0]["created"] is True, merge_response.responses

        def get_practitioner_graph(client_slug: str) -> Dict[str, Any]:
            return {
                "resourceType": "GraphDefinition",
                "id": "o",
                "name": "provider_everything",
                "status": "active",
                "start": "Practitioner",
                "link": [
                    {
                        "description": "Practitioner Roles for this Practitioner",
                        "target": [
                            {
                                "type": "PractitionerRole",
                                "params": f"practitioner={{ref}}&active:not=false&_security=https://www.icanbwell.com/access|{client_slug},https://www.icanbwell.com/access|connecthub",
                                "link": [
                                    {
                                        "path": "organization",
                                        "target": [
                                            {
                                                "type": "Organization",
                                                "link": [
                                                    {
                                                        "path": "endpoint[x]",
                                                        "target": [
                                                            {"type": "Endpoint"}
                                                        ],
                                                    }
                                                ],
                                            }
                                        ],
                                    },
                                    {
                                        "path": "location[x]",
                                        "target": [{"type": "Location"}],
                                    },
                                    {
                                        "path": "healthcareService[x]",
                                        "target": [{"type": "HealthcareService"}],
                                    },
                                    {
                                        "path": "extension.extension:url=plan",
                                        "target": [
                                            {
                                                "link": [
                                                    {
                                                        "path": "valueReference",
                                                        "target": [
                                                            {"type": "InsurancePlan"}
                                                        ],
                                                    }
                                                ]
                                            }
                                        ],
                                    },
                                    {
                                        "target": [
                                            {
                                                "type": "Schedule",
                                                "params": f"actor={{ref}}&_security=https://www.icanbwell.com/access|{client_slug}",
                                            }
                                        ]
                                    },
                                ],
                            }
                        ],
                    },
                    {
                        "description": "Group",
                        "target": [
                            {
                                "type": "Group",
                                "params": f"member={{ref}}&_security=https://www.icanbwell.com/access|{client_slug}",
                            }
                        ],
                    },
                    {
                        "description": "Review score for the practitioner",
                        "target": [
                            {
                                "type": "MeasureReport",
                                "params": f"subject={{ref}}&_security=https://www.icanbwell.com/access|{client_slug}",
                            }
                        ],
                    },
                ],
            }

        # act
        df: DataFrame = create_empty_dataframe(spark_session=spark_session)

        id_df = spark_session.createDataFrame(
            [(s,) for s in id_dict[resource_type]], ["id"]
        )

        id_df.createOrReplaceTempView("id_view")

        parameters = {"flow_name": "Test Pipeline V2", "team_name": "Data Operations"}

        with ProgressLogger() as progress_logger:
            await FhirReceiver(
                server_url=fhir_server_url,
                resource=resource_type,
                id_view="id_view",
                action="$graph",
                additional_parameters=["contained=true"],
                separate_bundle_resources=True,
                action_payload=get_practitioner_graph(client_slug="bwell"),
                file_path=patient_json_path,
                progress_logger=progress_logger,
                parameters=parameters,
                run_synchronously=run_synchronously,
                use_data_streaming=use_data_streaming,
                auth_well_known_url=auth_well_known_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
            ).transform_async(df)

            # Assert
            json_df: DataFrame = df.sparkSession.read.text(str(patient_json_path))
            print(" ---- Result of $graph call ------")
            json_df.show(truncate=False)
            print(" ---- Schema of $graph call ------")
            json_df.printSchema()

            assert json_df.count() == count

            assert json_df.schema == StructType(
                [
                    StructField("value", StringType(), True),
                ]
            )

            # now try to read it
            await FhirReader(
                file_path=patient_json_path,
                view="practitioner_graphs",
                name="fhir_reader",
                progress_logger=progress_logger,
                schema=StructType(
                    [
                        StructField(
                            "practitioner",
                            ArrayType(
                                PractitionerSchema.get_schema(
                                    include_extension=True,
                                )
                            ),
                        ),
                        StructField(
                            "practitionerrole",
                            ArrayType(
                                PractitionerRoleSchema.get_schema(
                                    include_extension=True,
                                    extension_fields=[
                                        "valueCodeableConcept",
                                        "valueRange",
                                        "valueReference",
                                    ],
                                )
                            ),
                        ),
                        StructField(
                            "organization",
                            ArrayType(OrganizationSchema.get_schema()),
                        ),
                        StructField(
                            "location",
                            ArrayType(
                                LocationSchema.get_schema(
                                    include_extension=True,
                                    extension_fields=["valueString"],
                                )
                            ),
                        ),
                        StructField(
                            "insuranceplan",
                            ArrayType(InsurancePlanSchema.get_schema()),
                        ),
                        StructField(
                            "endpoint",
                            ArrayType(
                                EndpointSchema.get_schema(
                                    include_extension=False,
                                )
                            ),
                        ),
                        StructField(
                            "schedule",
                            ArrayType(
                                ScheduleSchema.get_schema(
                                    include_extension=True,
                                    extension_fields=[
                                        "valueCodeableConcept",
                                        "valueRange",
                                        "valueDecimal",
                                    ],
                                )
                            ),
                        ),
                        StructField("group", ArrayType(GroupSchema.get_schema())),
                        StructField(
                            "measurereport", ArrayType(MeasureReportSchema.get_schema())
                        ),
                    ]
                ),
            ).transform_async(df)

            # Assert
            practitioner_graphs_df: DataFrame = spark_session.table(
                "practitioner_graphs"
            )
            print(" ---- Result of FhirReader ------")
            practitioner_graphs_df.show(truncate=False)
            print(" ---- Schema of FhirReader ------")
            practitioner_graphs_df.printSchema()

            assert practitioner_graphs_df.count() == count

            rows: List[Dict[str, Any]] = [
                r.asDict(recursive=True) for r in practitioner_graphs_df.collect()
            ]

            rows = sorted(rows, key=lambda x: x["practitioner"][0]["id"])

            assert rows[0]["practitioner"][0]["id"] == id_dict[resource_type][0]
            assert rows[1]["practitioner"][0]["id"] == id_dict[resource_type][1]
